# type: ignore
"""create postgres versioning triggers

Revision ID: fbda0eb90f37
Revises: ad25c5b542c4
Create Date: 2025-08-10 01:30:07.294959

"""

import logging

import sqlalchemy as sa
from alembic import op

from ixmp4.data.db.versions import PostgresVersionTriggers

logger = logging.getLogger(__name__)
# Revision identifiers, used by Alembic.
revision = "fbda0eb90f37"
down_revision = "ad25c5b542c4"
branch_labels = None
depends_on = None

transaction_table_name = "transaction"
tables = [
    ("iamc_datapoint_universal", "iamc_datapoint_universal_version"),
    ("iamc_measurand", "iamc_measurand_version"),
    ("iamc_timeseries", "iamc_timeseries_version"),
    ("iamc_variable", "iamc_variable_version"),
    ("model", "model_version"),
    ("region", "region_version"),
    ("run", "run_version"),
    ("runmetaentry", "runmetaentry_version"),
    ("scenario", "scenario_version"),
    ("unit", "unit_version"),
]


def make_versioning_triggers(
    table_name: str, version_table_name: str, transaction_table: sa.Table
):
    conn = op.get_bind()

    version_table = sa.Table(
        version_table_name,
        sa.MetaData(),
        autoload_with=conn,
    )

    data_table = sa.Table(
        table_name,
        sa.MetaData(),
        autoload_with=conn,
    )
    return PostgresVersionTriggers(data_table, version_table, transaction_table)


def upgrade():
    conn = op.get_bind()
    if conn is None:
        logging.warning("Cannot create version triggers without database connection.")
        return

    dialect_name = conn.dialect.name
    if dialect_name != "postgresql":
        logging.info(
            "Skipping creation of version triggers for "
            f"database dialect '{dialect_name}'."
        )
        return

    transaction_table = sa.Table(
        transaction_table_name,
        sa.MetaData(),
        autoload_with=conn,
    )
    for table, version_table in tables:
        triggers = make_versioning_triggers(table, version_table, transaction_table)
        triggers.create_entities(conn)


def downgrade():
    conn = op.get_bind()
    if conn is None:
        logging.warning("Cannot create version triggers without database connection.")
        return

    dialect_name = conn.dialect.name
    if dialect_name != "postgresql":
        logging.info(
            "Skipping dropping of version triggers for "
            f"database dialect '{dialect_name}'."
        )
        return

    transaction_table = sa.Table(
        transaction_table_name,
        sa.MetaData(),
        autoload_with=conn,
    )
    for table, version_table in tables:
        triggers = make_versioning_triggers(table, version_table, transaction_table)
        triggers.drop_entities(conn)



from collections.abc import Mapping
from typing import Any, ClassVar, cast

import pandas as pd

# TODO Import this from typing when dropping support for Python 3.11
from typing_extensions import Unpack

from ixmp4.data import abstract

from .. import base


class TimeSeries(base.BaseModel):
    NotFound: ClassVar = abstract.TimeSeries.NotFound
    NotUnique: ClassVar = abstract.TimeSeries.NotUnique
    DeletionPrevented: ClassVar = abstract.TimeSeries.DeletionPrevented

    id: int
    run__id: int
    parameters: Mapping[str, Any]


class TimeSeriesRepository(
    base.Creator[TimeSeries],
    base.Retriever[TimeSeries],
    base.Enumerator[TimeSeries],
    base.BulkUpserter[TimeSeries],
    abstract.TimeSeriesRepository[abstract.TimeSeries],
):
    model_class = TimeSeries
    prefix = "iamc/timeseries/"

    def create(self, run_id: int, parameters: Mapping[str, Any]) -> TimeSeries:
        return super().create(run_id=run_id, parameters=parameters)

    def get(self, run_id: int, parameters: Mapping[str, Any]) -> TimeSeries:
        return super().get(run_ids=[run_id], parameters=parameters)

    def enumerate(
        self, **kwargs: Unpack[abstract.iamc.timeseries.EnumerateKwargs]
    ) -> list[TimeSeries] | pd.DataFrame:
        return super().enumerate(**kwargs)

    def list(
        self, **kwargs: Unpack[abstract.iamc.timeseries.EnumerateKwargs]
    ) -> list[TimeSeries]:
        json = cast(abstract.annotations.IamcFilterAlias, kwargs)
        return super()._list(json=json)

    def tabulate(
        self,
        join_parameters: bool | None = None,
        **kwargs: Unpack[abstract.iamc.timeseries.EnumerateKwargs],
    ) -> pd.DataFrame:
        json = cast(abstract.annotations.IamcFilterAlias, kwargs)
        return super()._tabulate(json=json, params={"join_parameters": join_parameters})

    def get_by_id(self, id: int) -> TimeSeries:
        res = self._get_by_id(id)
        return TimeSeries(**res)

    def bulk_upsert(self, df: pd.DataFrame, create_related: bool = False) -> None:
        super().bulk_upsert(df, create_related=create_related)



from contextlib import contextmanager
from typing import Generator

from alembic import op
from sqlalchemy import text


@contextmanager
def pause_foreign_key_checks() -> Generator[None, None, None]:
    """
    Context manager to pause foreign key checks for the duration of the block.

    This is useful for performing operations that would otherwise violate
    foreign key constraints, such as dropping or altering tables.
    """
    context = op.get_context()

    if context.dialect.name == "sqlite":
        op.execute("PRAGMA foreign_keys = OFF;")
    elif context.dialect.name == "postgresql":
        # y i k e s
        tables = [
            row[0]
            for row in op.get_bind().execute(
                text("SELECT tablename FROM pg_tables WHERE schemaname = 'public';")
            )
        ]
        for table in tables:
            op.execute(f'ALTER TABLE "{table}" DISABLE TRIGGER USER;')
    else:
        raise NotImplementedError(
            "Foreign key checks cannot be paused for dialect "
            f"'{op.get_context().dialect.name}'"
        )

    try:
        yield
    finally:
        if context.dialect.name == "sqlite":
            op.execute("PRAGMA foreign_keys = ON;")
        elif context.dialect.name == "postgresql":
            tables = [
                row[0]
                for row in op.get_bind().execute(
                    text("SELECT tablename FROM pg_tables WHERE schemaname = 'public';")
                )
            ]
            for table in tables:
                op.execute(f'ALTER TABLE "{table}" ENABLE TRIGGER USER;')



from typing import ClassVar

from ixmp4.data.db.optimization.table.model import TableIndexsetAssociation
from ixmp4.db import Session, filters, sql

from .. import Run, Table


class OptimizationTableFilter(filters.BaseFilter, metaclass=filters.FilterMeta):
    id: filters.Id | None = filters.Field(None)
    name: filters.String | None = filters.Field(None)
    run__id: filters.Integer | None = filters.Field(None, alias="run_id")

    sqla_model: ClassVar[type] = Table

    def join(
        self, exc: sql.Select[tuple[Table]], session: Session | None = None
    ) -> sql.Select[tuple[Table]]:
        exc = exc.join(Run, onclause=Table.run__id == Run.id)
        return exc


class OptimizationTableIndexSetAssociationFilter(
    filters.BaseFilter, metaclass=filters.FilterMeta
):
    id: filters.Id | None = filters.Field(None)
    table__id: filters.Integer | None = filters.Field(None, alias="table_id")

    sqla_model: ClassVar[type] = TableIndexsetAssociation



import logging
from collections.abc import Iterable
from typing import TYPE_CHECKING, Any, cast

import pandas as pd

# TODO Import this from typing when dropping Python 3.11
from typing_extensions import Unpack

from ixmp4 import db
from ixmp4.core.exceptions import OptimizationItemUsageError
from ixmp4.data import types
from ixmp4.data.abstract import optimization as abstract
from ixmp4.data.auth.decorators import guard
from ixmp4.data.db import versions
from ixmp4.data.db.optimization.associations import (
    BaseIndexSetAssociationRepository,
    BaseIndexSetAssociationReverter,
    BaseIndexSetAssociationVersionRepository,
)

from .. import base
from .docs import EquationDocsRepository
from .model import (
    Equation,
    EquationIndexsetAssociation,
    EquationIndexsetAssociationVersion,
    EquationVersion,
)

if TYPE_CHECKING:
    from ixmp4.data.backend.db import SqlAlchemyBackend

logger = logging.getLogger(__name__)


class EquationVersionRepository(versions.VersionRepository[EquationVersion]):
    model_class = EquationVersion


class EquationIndexSetAssociationVersionRepository(
    BaseIndexSetAssociationVersionRepository[EquationIndexsetAssociationVersion]
):
    model_class = EquationIndexsetAssociationVersion
    parent_version = EquationVersion
    _item_id_column = "equation__id"


class EquationIndexSetAssociationRepository(
    BaseIndexSetAssociationRepository[
        EquationIndexsetAssociation, EquationIndexsetAssociationVersion
    ]
):
    model_class = EquationIndexsetAssociation
    versions: EquationIndexSetAssociationVersionRepository

    def __init__(self, backend: "SqlAlchemyBackend") -> None:
        super().__init__(backend)

        self.versions = EquationIndexSetAssociationVersionRepository(backend)

        from .filter import OptimizationEquationIndexSetAssociationFilter

        self.filter_class = OptimizationEquationIndexSetAssociationFilter


class EquationRepository(
    base.Creator[Equation],
    base.Deleter[Equation],
    base.Retriever[Equation],
    base.Enumerator[Equation],
    BaseIndexSetAssociationReverter[
        EquationIndexsetAssociation, EquationIndexsetAssociationVersion
    ],
    abstract.EquationRepository,
):
    model_class = Equation

    UsageError = OptimizationItemUsageError

    versions: EquationVersionRepository

    def __init__(self, *args: "SqlAlchemyBackend") -> None:
        super().__init__(*args)
        self.docs = EquationDocsRepository(*args)

        from .filter import EquationFilter

        self.filter_class = EquationFilter

        self.versions = EquationVersionRepository(*args)

        self._associations = EquationIndexSetAssociationRepository(*args)

        self._item_id_column = "equation__id"

    def add(
        self,
        run_id: int,
        name: str,
        constrained_to_indexsets: list[str] | None = None,
        column_names: list[str] | None = None,
    ) -> Equation:
        equation = Equation(name=name, run__id=run_id)

        if constrained_to_indexsets:
            indexsets = {
                indexset: self.backend.optimization.indexsets.get(
                    run_id=run_id, name=indexset
                )
                for indexset in set(constrained_to_indexsets)
            }
            for i in range(len(constrained_to_indexsets)):
                _ = EquationIndexsetAssociation(
                    equation=equation,
                    indexset=indexsets[constrained_to_indexsets[i]],
                    column_name=column_names[i] if column_names else None,
                )

        equation.set_creation_info(auth_context=self.backend.auth_context)
        self.session.add(equation)

        return equation

    @guard("view")
    def get(self, run_id: int, name: str) -> Equation:
        exc = db.select(Equation).where(
            (Equation.name == name) & (Equation.run__id == run_id)
        )
        try:
            return self.session.execute(exc).scalar_one()
        except db.NoResultFound:
            raise Equation.NotFound

    @guard("view")
    def get_by_id(self, id: int) -> Equation:
        obj = self.session.get(self.model_class, id)

        if obj is None:
            raise Equation.NotFound(id=id)

        return obj

    @guard("edit")
    def create(
        self,
        name: str,
        run_id: int,
        constrained_to_indexsets: list[str] | None = None,
        column_names: list[str] | None = None,
    ) -> Equation:
        if column_names:
            # TODO If this is removed, need to check above that constrained_to_indexsets
            #  is not None
            if constrained_to_indexsets is None:
                raise self.UsageError(
                    f"While processing Variable {name}: \n"
                    "Received `column_names` to name columns, but no "
                    "`constrained_to_indexsets` to indicate which IndexSets to use for "
                    "these columns. Please provide `constrained_to_indexsets` or "
                    "remove `column_names`!"
                )
            elif len(column_names) != len(constrained_to_indexsets):
                raise OptimizationItemUsageError(
                    f"While processing Equation {name}: \n"
                    "`constrained_to_indexsets` and `column_names` not equal in length!"
                    "Please provide the same number of entries for both!"
                )
            elif len(column_names) != len(set(column_names)):
                raise OptimizationItemUsageError(
                    f"While processing Equation {name}: \n"
                    "The given `column_names` are not unique!"
                )

        return super().create(
            run_id=run_id,
            name=name,
            constrained_to_indexsets=constrained_to_indexsets,
            column_names=column_names,
        )

    @guard("edit")
    def delete(self, id: int) -> None:
        # Manually ensure associations are deleted first to fix order of version updates
        associations_to_delete = self._associations.tabulate(equation_id=id)
        self._associations.bulk_delete(associations_to_delete)
        super().delete(id=id)

    @guard("view")
    def list(self, **kwargs: Unpack["base.EnumerateKwargs"]) -> Iterable[Equation]:
        return super().list(**kwargs)

    @guard("view")
    def tabulate(self, **kwargs: Unpack["base.EnumerateKwargs"]) -> pd.DataFrame:
        return super().tabulate(**kwargs)

    @guard("edit")
    def add_data(self, id: int, data: dict[str, Any] | pd.DataFrame) -> None:
        if isinstance(data, dict):
            try:
                data = pd.DataFrame.from_dict(data=data)
            except ValueError as e:
                raise Equation.DataInvalid(str(e)) from e

        if data.empty:
            return  # nothing to do

        equation = self.get_by_id(id=id)

        missing_columns = set(["levels", "marginals"]) - set(data.columns)
        if missing_columns:
            raise OptimizationItemUsageError(
                f"Equation.data must include the column(s): "
                f"{', '.join(missing_columns)}!"
            )

        index_list = equation.column_names or equation.indexset_names
        existing_data = pd.DataFrame(equation.data)
        if index_list:
            data = data.set_index(index_list)
            if not existing_data.empty:
                existing_data.set_index(index_list, inplace=True)
        data = data.combine_first(existing_data)
        if index_list:
            data = data.reset_index()

        equation.data = cast(types.JsonDict, data.to_dict(orient="list"))

        self.session.commit()

    @guard("edit")
    def remove_data(
        self, id: int, data: dict[str, Any] | pd.DataFrame | None = None
    ) -> None:
        equation = self.get_by_id(id=id)

        if data is None:
            # Remove all data per default
            # TODO Is there a better way to reset .data?
            equation.data = {}
        else:
            if isinstance(data, dict):
                data = pd.DataFrame.from_dict(data=data)

            if data.empty:
                return

            index_list = equation.column_names or equation.indexset_names
            if not index_list:
                logger.warning(
                    f"Trying to remove {data.to_dict(orient='list')} from Equation '"
                    f"{equation.name}', but that is not indexed; not removing anything!"
                )
                return  # can't remove specific data from unindexed variable

            existing_data = pd.DataFrame(equation.data)
            if not existing_data.empty:
                existing_data.set_index(index_list, inplace=True)

            # This is the only kind of validation we do for removal data
            try:
                data.set_index(index_list, inplace=True)
            except KeyError as e:
                logger.error(
                    f"Data to be removed must include {index_list} as keys/columns, "
                    f"but {[name for name in data.columns]} were provided."
                )
                raise OptimizationItemUsageError(
                    "The data to be removed must specify one or more complete indices "
                    "to remove associated levels and marginals!"
                ) from e

            remaining_data = existing_data[~existing_data.index.isin(data.index)]
            if not remaining_data.index.empty:
                remaining_data.reset_index(inplace=True)

            equation.data = cast(types.JsonDict, remaining_data.to_dict(orient="list"))

        self.session.commit()

    @guard("edit")
    def revert(self, transaction__id: int, run__id: int) -> None:
        super().revert(transaction__id=transaction__id, run__id=run__id)

        # Revert EquationIndexSetAssociation
        self._revert_indexset_association(
            repo=self, transaction__id=transaction__id, run__id=run__id
        )



from collections.abc import Iterable
from datetime import datetime

import pandas as pd

from ixmp4.core.base import BaseFacade, BaseModelFacade
from ixmp4.data.abstract import Docs as DocsModel
from ixmp4.data.abstract import Unit as UnitModel


class Unit(BaseModelFacade):
    _model: UnitModel
    NotUnique = UnitModel.NotUnique
    NotFound = UnitModel.NotFound
    DeletionPrevented = UnitModel.DeletionPrevented

    @property
    def id(self) -> int:
        return self._model.id

    @property
    def name(self) -> str:
        return self._model.name

    @property
    def created_at(self) -> datetime | None:
        return self._model.created_at

    @property
    def created_by(self) -> str | None:
        return self._model.created_by

    def delete(self) -> None:
        self.backend.units.delete(self._model.id)

    @property
    def docs(self) -> str | None:
        try:
            return self.backend.units.docs.get(self.id).description
        except DocsModel.NotFound:
            return None

    @docs.setter
    def docs(self, description: str | None) -> None:
        if description is None:
            self.backend.units.docs.delete(self.id)
        else:
            self.backend.units.docs.set(self.id, description)

    @docs.deleter
    def docs(self) -> None:
        try:
            self.backend.units.docs.delete(self.id)
        # TODO: silently failing
        except DocsModel.NotFound:
            return None

    def __str__(self) -> str:
        return f"<Unit {self.id} name={self.name}>"


class UnitRepository(BaseFacade):
    def create(self, name: str) -> Unit:
        if name != "" and name.strip() == "":
            raise ValueError("Using a space-only unit name is not allowed.")
        if name == "dimensionless":
            raise ValueError(
                "Unit name 'dimensionless' is reserved, use an empty string '' instead."
            )
        model = self.backend.units.create(name)
        return Unit(_backend=self.backend, _model=model)

    def delete(self, x: Unit | int | str) -> None:
        if isinstance(x, Unit):
            id = x.id
        elif isinstance(x, int):
            id = x
        elif isinstance(x, str):
            model = self.backend.units.get(x)
            id = model.id
        else:
            raise TypeError("Invalid argument: Must be `Unit`, `int` or `str`.")

        self.backend.units.delete(id)

    def get(self, name: str) -> Unit:
        model = self.backend.units.get(name)
        return Unit(_backend=self.backend, _model=model)

    def list(self, name: str | None = None) -> list[Unit]:
        units = self.backend.units.list(name=name)
        return [Unit(_backend=self.backend, _model=u) for u in units]

    def tabulate(self, name: str | None = None) -> pd.DataFrame:
        return self.backend.units.tabulate(name=name)

    def _get_unit_id(self, unit: str) -> int | None:
        # NOTE leaving this check for users without mypy
        if isinstance(unit, str):
            obj = self.backend.units.get(unit)
            return obj.id
        else:
            raise ValueError(f"Invalid reference to unit: {unit}")

    def get_docs(self, name: str) -> str | None:
        unit_id = self._get_unit_id(name)
        if unit_id is None:
            return None
        try:
            return self.backend.units.docs.get(dimension_id=unit_id).description
        except DocsModel.NotFound:
            return None

    def set_docs(self, name: str, description: str | None) -> str | None:
        if description is None:
            self.delete_docs(name=name)
            return None
        unit_id = self._get_unit_id(name)
        if unit_id is None:
            return None
        return self.backend.units.docs.set(
            dimension_id=unit_id, description=description
        ).description

    def delete_docs(self, name: str) -> None:
        # TODO: this function is failing silently, which we should avoid
        unit_id = self._get_unit_id(name)
        if unit_id is None:
            return None
        try:
            self.backend.units.docs.delete(dimension_id=unit_id)
            return None
        except DocsModel.NotFound:
            return None

    def list_docs(
        self, id: int | None = None, id__in: Iterable[int] | None = None
    ) -> Iterable[str]:
        return [
            item.description
            for item in self.backend.units.docs.list(
                dimension_id=id, dimension_id__in=id__in
            )
        ]



from typing import Any

from ixmp4.data.db.docs import BaseDocsRepository, docs_model

from .model import Table

TableDocs = docs_model(Table)


class TableDocsRepository(BaseDocsRepository[Any]):
    model_class = TableDocs
    dimension_model_class = Table



from typing import ClassVar

from ixmp4.data.db.optimization.parameter.model import ParameterIndexsetAssociation
from ixmp4.db import Session, filters, sql

from .. import Parameter, Run


class OptimizationParameterFilter(filters.BaseFilter, metaclass=filters.FilterMeta):
    id: filters.Id | None = filters.Field(None)
    name: filters.String | None = filters.Field(None)
    run__id: filters.Integer | None = filters.Field(None, alias="run_id")

    sqla_model: ClassVar[type] = Parameter

    def join(
        self, exc: sql.Select[tuple[Parameter]], session: Session | None = None
    ) -> sql.Select[tuple[Parameter]]:
        exc = exc.join(Run, onclause=Parameter.run__id == Run.id)
        return exc


class OptimizationParameterIndexSetAssociationFilter(
    filters.BaseFilter, metaclass=filters.FilterMeta
):
    id: filters.Id | None = filters.Field(None)
    parameter__id: filters.Integer | None = filters.Field(None, alias="parameter_id")

    sqla_model: ClassVar[type] = ParameterIndexsetAssociation



"""
When making schema migrations for versioned tables
you need to remember to call `sync_trigger` in order
to keep the version trigger up-to-date.

```
from alembic import context

if not context.is_offline_mode():
    conn = context.get_bind()
    sync_trigger(conn, "my_table_version")
```
"""

from typing import Literal

from alembic import context
from sqlalchemy import Engine, create_engine
from sqlalchemy.orm import configure_mappers
from sqlalchemy.schema import SchemaItem

from ixmp4.conf import settings
from ixmp4.data.db import BaseModel

# this is the Alembic Config object, which provides
# access to the values within the .ini file in use.
config = context.config
dsn = config.get_main_option("sqlalchemy.url", settings.migration_db_uri)
dsn = dsn.replace("postgresql://", "postgresql+psycopg://")

configure_mappers()

target_metadata = BaseModel.metadata

# other values from the config, defined by the needs of env.py,
# can be acquired:
# my_important_option = config.get_main_option("my_important_option")
# ... etc.


def include_object(
    obj: SchemaItem,
    name: str | None,
    type_: Literal[
        "schema",
        "table",
        "column",
        "index",
        "unique_constraint",
        "foreign_key_constraint",
    ],
    reflected: bool,
    compare_to: SchemaItem | None,
) -> bool:
    if type_ == "column" and not reflected and obj.info.get("skip_autogenerate", False):
        return False
    else:
        return True


def run_migrations_offline() -> None:
    """Run migrations in 'offline' mode.

    This configures the context with just a URL
    and not an Engine, though an Engine is acceptable
    here as well.  By skipping the Engine creation
    we don't even need a DBAPI to be available.

    Calls to context.execute() here emit the given string to the
    script output.

    """
    context.configure(
        url=dsn,
        target_metadata=target_metadata,
        compare_type=True,
        literal_binds=True,
        render_as_batch=True,
        include_object=include_object,
        dialect_opts={"paramstyle": "named"},
    )

    with context.begin_transaction():
        context.run_migrations()


def run_migrations_online() -> None:
    """Run migrations in 'online' mode.

    In this scenario we need to create an Engine
    and associate a connection with the context.

    """
    engine: Engine | None = context.config.attributes.get("connection", None)
    if engine is None:
        engine = create_engine(dsn, max_identifier_length=63)

    with engine.connect() as conn:
        context.configure(
            connection=conn,
            target_metadata=target_metadata,
            compare_type=True,
            render_as_batch=True,
            include_object=include_object,
        )

        with context.begin_transaction():
            context.run_migrations()


if context.is_offline_mode():
    run_migrations_offline()
else:
    run_migrations_online()



# type: ignore
"""add version tables

Revision ID: 0b45de1602a6
Revises: d66a8276ba0a
Create Date: 2025-01-14 11:18:19.196147

"""

import sqlalchemy as sa
from alembic import op

# Revision identifiers, used by Alembic.
revision = "0b45de1602a6"
down_revision = "d66a8276ba0a"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "iamc_datapoint_universal_version",
        sa.Column("value", sa.Float(), autoincrement=False, nullable=True),
        sa.Column("type", sa.String(length=255), autoincrement=False, nullable=True),
        sa.Column(
            "step_category", sa.String(length=1023), autoincrement=False, nullable=True
        ),
        sa.Column("step_year", sa.Integer(), autoincrement=False, nullable=True),
        sa.Column("step_datetime", sa.DateTime(), autoincrement=False, nullable=True),
        sa.Column("time_series__id", sa.Integer(), autoincrement=False, nullable=True),
        sa.Column("id", sa.Integer(), autoincrement=False, nullable=False),
        sa.Column(
            "transaction_id", sa.BigInteger(), autoincrement=False, nullable=False
        ),
        sa.Column("end_transaction_id", sa.BigInteger(), nullable=True),
        sa.Column("operation_type", sa.SmallInteger(), nullable=False),
        sa.PrimaryKeyConstraint(
            "id", "transaction_id", name=op.f("pk_iamc_datapoint_universal_version")
        ),
    )
    with op.batch_alter_table(
        "iamc_datapoint_universal_version", schema=None
    ) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_iamc_datapoint_universal_version_end_transaction_id"),
            ["end_transaction_id"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_iamc_datapoint_universal_version_operation_type"),
            ["operation_type"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_iamc_datapoint_universal_version_step_category"),
            ["step_category"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_iamc_datapoint_universal_version_step_datetime"),
            ["step_datetime"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_iamc_datapoint_universal_version_step_year"),
            ["step_year"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_iamc_datapoint_universal_version_time_series__id"),
            ["time_series__id"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_iamc_datapoint_universal_version_transaction_id"),
            ["transaction_id"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_iamc_datapoint_universal_version_type"),
            ["type"],
            unique=False,
        )

    op.create_table(
        "iamc_measurand_version",
        sa.Column("variable__id", sa.Integer(), autoincrement=False, nullable=True),
        sa.Column("unit__id", sa.Integer(), autoincrement=False, nullable=True),
        sa.Column("id", sa.Integer(), autoincrement=False, nullable=False),
        sa.Column("created_at", sa.DateTime(), autoincrement=False, nullable=True),
        sa.Column(
            "created_by", sa.String(length=255), autoincrement=False, nullable=True
        ),
        sa.Column(
            "transaction_id", sa.BigInteger(), autoincrement=False, nullable=False
        ),
        sa.Column("end_transaction_id", sa.BigInteger(), nullable=True),
        sa.Column("operation_type", sa.SmallInteger(), nullable=False),
        sa.PrimaryKeyConstraint(
            "id", "transaction_id", name=op.f("pk_iamc_measurand_version")
        ),
    )
    with op.batch_alter_table("iamc_measurand_version", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_iamc_measurand_version_end_transaction_id"),
            ["end_transaction_id"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_iamc_measurand_version_operation_type"),
            ["operation_type"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_iamc_measurand_version_transaction_id"),
            ["transaction_id"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_iamc_measurand_version_unit__id"), ["unit__id"], unique=False
        )
        batch_op.create_index(
            batch_op.f("ix_iamc_measurand_version_variable__id"),
            ["variable__id"],
            unique=False,
        )

    op.create_table(
        "iamc_timeseries_version",
        sa.Column("region__id", sa.Integer(), autoincrement=False, nullable=True),
        sa.Column("measurand__id", sa.Integer(), autoincrement=False, nullable=True),
        sa.Column("run__id", sa.Integer(), autoincrement=False, nullable=True),
        sa.Column("id", sa.Integer(), autoincrement=False, nullable=False),
        sa.Column(
            "transaction_id", sa.BigInteger(), autoincrement=False, nullable=False
        ),
        sa.Column("end_transaction_id", sa.BigInteger(), nullable=True),
        sa.Column("operation_type", sa.SmallInteger(), nullable=False),
        sa.PrimaryKeyConstraint(
            "id", "transaction_id", name=op.f("pk_iamc_timeseries_version")
        ),
    )
    with op.batch_alter_table("iamc_timeseries_version", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_iamc_timeseries_version_end_transaction_id"),
            ["end_transaction_id"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_iamc_timeseries_version_measurand__id"),
            ["measurand__id"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_iamc_timeseries_version_operation_type"),
            ["operation_type"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_iamc_timeseries_version_region__id"),
            ["region__id"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_iamc_timeseries_version_run__id"), ["run__id"], unique=False
        )
        batch_op.create_index(
            batch_op.f("ix_iamc_timeseries_version_transaction_id"),
            ["transaction_id"],
            unique=False,
        )

    op.create_table(
        "iamc_variable_version",
        sa.Column("name", sa.String(length=255), autoincrement=False, nullable=True),
        sa.Column("id", sa.Integer(), autoincrement=False, nullable=False),
        sa.Column("created_at", sa.DateTime(), autoincrement=False, nullable=True),
        sa.Column(
            "created_by", sa.String(length=255), autoincrement=False, nullable=True
        ),
        sa.Column(
            "transaction_id", sa.BigInteger(), autoincrement=False, nullable=False
        ),
        sa.Column("end_transaction_id", sa.BigInteger(), nullable=True),
        sa.Column("operation_type", sa.SmallInteger(), nullable=False),
        sa.PrimaryKeyConstraint(
            "id", "transaction_id", name=op.f("pk_iamc_variable_version")
        ),
    )
    with op.batch_alter_table("iamc_variable_version", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_iamc_variable_version_end_transaction_id"),
            ["end_transaction_id"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_iamc_variable_version_operation_type"),
            ["operation_type"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_iamc_variable_version_transaction_id"),
            ["transaction_id"],
            unique=False,
        )

    op.create_table(
        "model_version",
        sa.Column("name", sa.String(length=255), autoincrement=False, nullable=True),
        sa.Column("id", sa.Integer(), autoincrement=False, nullable=False),
        sa.Column("created_at", sa.DateTime(), autoincrement=False, nullable=True),
        sa.Column(
            "created_by", sa.String(length=255), autoincrement=False, nullable=True
        ),
        sa.Column(
            "transaction_id", sa.BigInteger(), autoincrement=False, nullable=False
        ),
        sa.Column("end_transaction_id", sa.BigInteger(), nullable=True),
        sa.Column("operation_type", sa.SmallInteger(), nullable=False),
        sa.PrimaryKeyConstraint("id", "transaction_id", name=op.f("pk_model_version")),
    )
    with op.batch_alter_table("model_version", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_model_version_end_transaction_id"),
            ["end_transaction_id"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_model_version_operation_type"),
            ["operation_type"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_model_version_transaction_id"),
            ["transaction_id"],
            unique=False,
        )

    op.create_table(
        "region_version",
        sa.Column("name", sa.String(length=255), autoincrement=False, nullable=True),
        sa.Column(
            "hierarchy", sa.String(length=1023), autoincrement=False, nullable=True
        ),
        sa.Column("id", sa.Integer(), autoincrement=False, nullable=False),
        sa.Column("created_at", sa.DateTime(), autoincrement=False, nullable=True),
        sa.Column(
            "created_by", sa.String(length=255), autoincrement=False, nullable=True
        ),
        sa.Column(
            "transaction_id", sa.BigInteger(), autoincrement=False, nullable=False
        ),
        sa.Column("end_transaction_id", sa.BigInteger(), nullable=True),
        sa.Column("operation_type", sa.SmallInteger(), nullable=False),
        sa.PrimaryKeyConstraint("id", "transaction_id", name=op.f("pk_region_version")),
    )
    with op.batch_alter_table("region_version", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_region_version_end_transaction_id"),
            ["end_transaction_id"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_region_version_operation_type"),
            ["operation_type"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_region_version_transaction_id"),
            ["transaction_id"],
            unique=False,
        )

    op.create_table(
        "run_version",
        sa.Column("model__id", sa.Integer(), autoincrement=False, nullable=True),
        sa.Column("scenario__id", sa.Integer(), autoincrement=False, nullable=True),
        sa.Column("version", sa.Integer(), autoincrement=False, nullable=True),
        sa.Column("is_default", sa.Boolean(), autoincrement=False, nullable=True),
        sa.Column("id", sa.Integer(), autoincrement=False, nullable=False),
        sa.Column("updated_at", sa.DateTime(), autoincrement=False, nullable=True),
        sa.Column(
            "updated_by", sa.String(length=255), autoincrement=False, nullable=True
        ),
        sa.Column("created_at", sa.DateTime(), autoincrement=False, nullable=True),
        sa.Column(
            "created_by", sa.String(length=255), autoincrement=False, nullable=True
        ),
        sa.Column(
            "transaction_id", sa.BigInteger(), autoincrement=False, nullable=False
        ),
        sa.Column("end_transaction_id", sa.BigInteger(), nullable=True),
        sa.Column("operation_type", sa.SmallInteger(), nullable=False),
        sa.PrimaryKeyConstraint("id", "transaction_id", name=op.f("pk_run_version")),
    )
    with op.batch_alter_table("run_version", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_run_version_end_transaction_id"),
            ["end_transaction_id"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_run_version_model__id"), ["model__id"], unique=False
        )
        batch_op.create_index(
            batch_op.f("ix_run_version_operation_type"),
            ["operation_type"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_run_version_scenario__id"), ["scenario__id"], unique=False
        )
        batch_op.create_index(
            batch_op.f("ix_run_version_transaction_id"),
            ["transaction_id"],
            unique=False,
        )

    op.create_table(
        "runmetaentry_version",
        sa.Column("run__id", sa.Integer(), autoincrement=False, nullable=True),
        sa.Column("key", sa.String(length=1023), autoincrement=False, nullable=True),
        sa.Column("dtype", sa.String(length=20), autoincrement=False, nullable=True),
        sa.Column("value_int", sa.Integer(), autoincrement=False, nullable=True),
        sa.Column(
            "value_str", sa.String(length=1023), autoincrement=False, nullable=True
        ),
        sa.Column("value_float", sa.Float(), autoincrement=False, nullable=True),
        sa.Column("value_bool", sa.Boolean(), autoincrement=False, nullable=True),
        sa.Column("id", sa.Integer(), autoincrement=False, nullable=False),
        sa.Column(
            "transaction_id", sa.BigInteger(), autoincrement=False, nullable=False
        ),
        sa.Column("end_transaction_id", sa.BigInteger(), nullable=True),
        sa.Column("operation_type", sa.SmallInteger(), nullable=False),
        sa.PrimaryKeyConstraint(
            "id", "transaction_id", name=op.f("pk_runmetaentry_version")
        ),
    )
    with op.batch_alter_table("runmetaentry_version", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_runmetaentry_version_end_transaction_id"),
            ["end_transaction_id"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_runmetaentry_version_operation_type"),
            ["operation_type"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_runmetaentry_version_run__id"), ["run__id"], unique=False
        )
        batch_op.create_index(
            batch_op.f("ix_runmetaentry_version_transaction_id"),
            ["transaction_id"],
            unique=False,
        )

    op.create_table(
        "scenario_version",
        sa.Column("name", sa.String(length=255), autoincrement=False, nullable=True),
        sa.Column("id", sa.Integer(), autoincrement=False, nullable=False),
        sa.Column("created_at", sa.DateTime(), autoincrement=False, nullable=True),
        sa.Column(
            "created_by", sa.String(length=255), autoincrement=False, nullable=True
        ),
        sa.Column(
            "transaction_id", sa.BigInteger(), autoincrement=False, nullable=False
        ),
        sa.Column("end_transaction_id", sa.BigInteger(), nullable=True),
        sa.Column("operation_type", sa.SmallInteger(), nullable=False),
        sa.PrimaryKeyConstraint(
            "id", "transaction_id", name=op.f("pk_scenario_version")
        ),
    )
    with op.batch_alter_table("scenario_version", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_scenario_version_end_transaction_id"),
            ["end_transaction_id"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_scenario_version_operation_type"),
            ["operation_type"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_scenario_version_transaction_id"),
            ["transaction_id"],
            unique=False,
        )

    op.create_table(
        "transaction",
        sa.Column(
            "id",
            sa.BigInteger(),
            sa.Identity(always=False, on_null=True, start=2, increment=1),
            nullable=False,
        ),
        sa.Column("remote_addr", sa.String(length=50), nullable=True),
        sa.Column("issued_at", sa.DateTime(), nullable=True),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_transaction")),
    )
    op.create_table(
        "unit_version",
        sa.Column("name", sa.String(length=255), autoincrement=False, nullable=True),
        sa.Column("id", sa.Integer(), autoincrement=False, nullable=False),
        sa.Column("created_at", sa.DateTime(), autoincrement=False, nullable=True),
        sa.Column(
            "created_by", sa.String(length=255), autoincrement=False, nullable=True
        ),
        sa.Column(
            "transaction_id", sa.BigInteger(), autoincrement=False, nullable=False
        ),
        sa.Column("end_transaction_id", sa.BigInteger(), nullable=True),
        sa.Column("operation_type", sa.SmallInteger(), nullable=False),
        sa.PrimaryKeyConstraint("id", "transaction_id", name=op.f("pk_unit_version")),
    )
    with op.batch_alter_table("unit_version", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_unit_version_end_transaction_id"),
            ["end_transaction_id"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_unit_version_operation_type"),
            ["operation_type"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_unit_version_transaction_id"),
            ["transaction_id"],
            unique=False,
        )

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("unit_version", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_unit_version_transaction_id"))
        batch_op.drop_index(batch_op.f("ix_unit_version_operation_type"))
        batch_op.drop_index(batch_op.f("ix_unit_version_end_transaction_id"))

    op.drop_table("unit_version")
    op.drop_table("transaction")
    with op.batch_alter_table("scenario_version", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_scenario_version_transaction_id"))
        batch_op.drop_index(batch_op.f("ix_scenario_version_operation_type"))
        batch_op.drop_index(batch_op.f("ix_scenario_version_end_transaction_id"))

    op.drop_table("scenario_version")
    with op.batch_alter_table("runmetaentry_version", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_runmetaentry_version_transaction_id"))
        batch_op.drop_index(batch_op.f("ix_runmetaentry_version_run__id"))
        batch_op.drop_index(batch_op.f("ix_runmetaentry_version_operation_type"))
        batch_op.drop_index(batch_op.f("ix_runmetaentry_version_end_transaction_id"))

    op.drop_table("runmetaentry_version")
    with op.batch_alter_table("run_version", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_run_version_transaction_id"))
        batch_op.drop_index(batch_op.f("ix_run_version_scenario__id"))
        batch_op.drop_index(batch_op.f("ix_run_version_operation_type"))
        batch_op.drop_index(batch_op.f("ix_run_version_model__id"))
        batch_op.drop_index(batch_op.f("ix_run_version_end_transaction_id"))

    op.drop_table("run_version")
    with op.batch_alter_table("region_version", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_region_version_transaction_id"))
        batch_op.drop_index(batch_op.f("ix_region_version_operation_type"))
        batch_op.drop_index(batch_op.f("ix_region_version_end_transaction_id"))

    op.drop_table("region_version")
    with op.batch_alter_table("model_version", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_model_version_transaction_id"))
        batch_op.drop_index(batch_op.f("ix_model_version_operation_type"))
        batch_op.drop_index(batch_op.f("ix_model_version_end_transaction_id"))

    op.drop_table("model_version")
    with op.batch_alter_table("iamc_variable_version", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_iamc_variable_version_transaction_id"))
        batch_op.drop_index(batch_op.f("ix_iamc_variable_version_operation_type"))
        batch_op.drop_index(batch_op.f("ix_iamc_variable_version_end_transaction_id"))

    op.drop_table("iamc_variable_version")
    with op.batch_alter_table("iamc_timeseries_version", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_iamc_timeseries_version_transaction_id"))
        batch_op.drop_index(batch_op.f("ix_iamc_timeseries_version_run__id"))
        batch_op.drop_index(batch_op.f("ix_iamc_timeseries_version_region__id"))
        batch_op.drop_index(batch_op.f("ix_iamc_timeseries_version_operation_type"))
        batch_op.drop_index(batch_op.f("ix_iamc_timeseries_version_measurand__id"))
        batch_op.drop_index(batch_op.f("ix_iamc_timeseries_version_end_transaction_id"))

    op.drop_table("iamc_timeseries_version")
    with op.batch_alter_table("iamc_measurand_version", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_iamc_measurand_version_variable__id"))
        batch_op.drop_index(batch_op.f("ix_iamc_measurand_version_unit__id"))
        batch_op.drop_index(batch_op.f("ix_iamc_measurand_version_transaction_id"))
        batch_op.drop_index(batch_op.f("ix_iamc_measurand_version_operation_type"))
        batch_op.drop_index(batch_op.f("ix_iamc_measurand_version_end_transaction_id"))

    op.drop_table("iamc_measurand_version")
    with op.batch_alter_table(
        "iamc_datapoint_universal_version", schema=None
    ) as batch_op:
        batch_op.drop_index(batch_op.f("ix_iamc_datapoint_universal_version_type"))
        batch_op.drop_index(
            batch_op.f("ix_iamc_datapoint_universal_version_transaction_id")
        )
        batch_op.drop_index(
            batch_op.f("ix_iamc_datapoint_universal_version_time_series__id")
        )
        batch_op.drop_index(batch_op.f("ix_iamc_datapoint_universal_version_step_year"))
        batch_op.drop_index(
            batch_op.f("ix_iamc_datapoint_universal_version_step_datetime")
        )
        batch_op.drop_index(
            batch_op.f("ix_iamc_datapoint_universal_version_step_category")
        )
        batch_op.drop_index(
            batch_op.f("ix_iamc_datapoint_universal_version_operation_type")
        )
        batch_op.drop_index(
            batch_op.f("ix_iamc_datapoint_universal_version_end_transaction_id")
        )

    op.drop_table("iamc_datapoint_universal_version")
    # ### end Alembic commands ###



from typing import TYPE_CHECKING, Protocol

if TYPE_CHECKING:
    from .. import Unit, Variable

import pandas as pd

from .. import base


class Measurand(base.BaseModel, Protocol):
    """Measurand data model."""

    variable__id: int
    "Foreign unique integer id of a variable."
    variable: "Variable"
    "Associated variable."

    unit__id: int
    "Foreign unique integer id of a unit."
    unit: "Unit"
    "Associated unit."

    def __str__(self) -> str:
        return f"<Measurand {self.id}>"


class MeasurandRepository(
    base.Creator,
    base.Retriever,
    base.Enumerator,
    Protocol,
):
    def create(self, variable_name: str, unit__id: int) -> Measurand: ...

    def get(self, variable_name: str, unit__id: int) -> Measurand: ...

    def get_or_create(self, variable_name: str, unit__id: int) -> Measurand:
        try:
            return self.get(variable_name, unit__id)
        except Measurand.NotFound:
            return self.create(variable_name, unit__id)

    def list(self) -> list[Measurand]: ...

    def tabulate(self) -> pd.DataFrame: ...



from datetime import datetime
from typing import TYPE_CHECKING, ClassVar, List, cast

import pandas as pd
from pydantic import StrictFloat, StrictInt, StrictStr

# TODO Import this from typing when dropping support for Python 3.11
from typing_extensions import Unpack

from ixmp4.data import abstract

from .. import base
from ..docs import Docs, DocsRepository

if TYPE_CHECKING:
    from ixmp4.data.backend.api import RestBackend


class IndexSet(base.BaseModel):
    NotFound: ClassVar = abstract.IndexSet.NotFound
    NotUnique: ClassVar = abstract.IndexSet.NotUnique
    DeletionPrevented: ClassVar = abstract.IndexSet.DeletionPrevented

    id: int
    name: str
    data: float | int | str | list[int] | list[float] | list[str] | None
    run__id: int

    created_at: datetime | None
    created_by: str | None


class IndexSetDocsRepository(DocsRepository):
    model_class = Docs
    prefix = "docs/optimization/indexsets/"


class IndexSetRepository(
    base.Creator[IndexSet],
    base.Deleter[IndexSet],
    base.Retriever[IndexSet],
    base.Enumerator[IndexSet],
    abstract.IndexSetRepository,
):
    model_class = IndexSet
    prefix = "optimization/indexsets/"

    def __init__(self, *args: Unpack[tuple["RestBackend"]]) -> None:
        super().__init__(*args)
        self.docs = IndexSetDocsRepository(self.backend)

    def create(self, run_id: int, name: str) -> IndexSet:
        return super().create(run_id=run_id, name=name)

    def delete(self, id: int) -> None:
        super().delete(id=id)

    def get(self, run_id: int, name: str) -> IndexSet:
        return super().get(name=name, run_id=run_id)

    def enumerate(
        self, **kwargs: Unpack[abstract.optimization.EnumerateKwargs]
    ) -> list[IndexSet] | pd.DataFrame:
        return super().enumerate(**kwargs)

    def list(
        self, **kwargs: Unpack[abstract.optimization.EnumerateKwargs]
    ) -> list[IndexSet]:
        json = cast(abstract.annotations.OptimizationFilterAlias, kwargs)
        return super()._list(json=json)

    def tabulate(
        self, **kwargs: Unpack[abstract.optimization.EnumerateKwargs]
    ) -> pd.DataFrame:
        json = cast(abstract.annotations.OptimizationFilterAlias, kwargs)
        return super()._tabulate(json=json)

    def add_data(
        self,
        id: int,
        data: StrictFloat
        | StrictInt
        | StrictStr
        | List[StrictFloat]
        | List[StrictInt]
        | List[StrictStr],
    ) -> None:
        kwargs = {"id": id, "data": data}
        self._request("PATCH", self.prefix + str(id) + "/data/", json=kwargs)

    def remove_data(
        self,
        id: int,
        data: StrictFloat
        | StrictInt
        | StrictStr
        | List[StrictFloat]
        | List[StrictInt]
        | List[StrictStr],
        remove_dependent_data: bool = True,
    ) -> None:
        kwargs = {"id": id, "data": data}
        self._request(
            "DELETE",
            self.prefix + str(id) + "/data/",
            params={"remove_dependent_data": remove_dependent_data},
            json=kwargs,
        )



from pathlib import Path

from ixmp4.conf import settings


def get_database_dir() -> Path:
    """Returns the path to the local sqlite database directory."""
    return settings.storage_directory / "databases"


def get_database_path(name: str) -> Path:
    """Returns a :class:`Path` object for a given sqlite database name.
    Does not check whether or not the file actually exists."""

    file_name = name + ".sqlite3"
    return get_database_dir() / file_name


def get_dsn(database_path: Path) -> str:
    """Returns sqlalchemy-friendly sqlite database URI for a given database name."""

    return database_path.absolute().as_uri().replace("file://", "sqlite:///")


def search_databases(name: str) -> str | None:
    """Returns a database URI if the desired database exists, otherwise `None`."""

    database_path = get_database_path(name)
    return get_dsn(database_path) if database_path.exists() else None



from typing import TYPE_CHECKING, Generic, TypeVar

import pandas as pd

# TODO Import this from typing when dropping Python 3.11
from typing_extensions import TypedDict, Unpack

from ixmp4.core.base import BaseFacade, BaseModelFacade
from ixmp4.data import abstract
from ixmp4.data.backend import Backend

if TYPE_CHECKING:
    from ixmp4.core.run import Run

    from . import InitKwargs


class BaseModelFacadeKwargs(TypedDict, total=False):
    _model: abstract.BaseModel | None
    _backend: Backend | None


class OptimizationBaseModelFacade(BaseModelFacade):
    _run: "Run"

    def __init__(self, _run: "Run", **kwargs: Unpack[BaseModelFacadeKwargs]) -> None:
        """Initialize an optimization item instance.

        Parameters
        ----------
        _run : `Run`
            The :class:`ixmp4.core.run.Run` this item belongs to.
        """
        # NOTE This function enables requiring a Run's lock to add or remove data to a
        # single optimization object. It also allows sidestepping the requirement to
        # have a Run locked before editing it. Thus, this function should only be called
        # internally.
        super().__init__(**kwargs)

        self._run = _run


FacadeOptimizationModelType = TypeVar(
    "FacadeOptimizationModelType", bound=OptimizationBaseModelFacade
)
AbstractOptimizationModelType = TypeVar(
    "AbstractOptimizationModelType", bound=abstract.BaseModel
)


class OptimizationBaseRepository(
    BaseFacade, Generic[FacadeOptimizationModelType, AbstractOptimizationModelType]
):
    _run: "Run"
    _backend_repository: abstract.BackendBaseRepository[AbstractOptimizationModelType]
    _model_type: type[FacadeOptimizationModelType]

    def __init__(self, _run: "Run", **kwargs: Unpack["InitKwargs"]) -> None:
        super().__init__(**kwargs)
        self._run = _run


class Creator(
    OptimizationBaseRepository[
        FacadeOptimizationModelType, AbstractOptimizationModelType
    ],
    abstract.Creator,
):
    def create(
        self, name: str, **kwargs: Unpack["abstract.optimization.base.CreateKwargs"]
    ) -> FacadeOptimizationModelType:
        self._run.require_lock()
        model = self._backend_repository.create(
            run_id=self._run.id, name=name, **kwargs
        )
        return self._model_type(_backend=self.backend, _model=model, _run=self._run)


class Deleter(
    OptimizationBaseRepository[
        FacadeOptimizationModelType, AbstractOptimizationModelType
    ],
    abstract.Deleter,
):
    def delete(self, item: int | str) -> None:
        self._run.require_lock()
        if isinstance(item, int):
            id = item
        elif isinstance(item, str):
            model = self._backend_repository.get(run_id=self._run.id, name=item)
            id = model.id
        else:
            raise TypeError("Invalid argument: `id` must be `int` or `str`.")

        self._backend_repository.delete(id=id)


class Retriever(
    OptimizationBaseRepository[
        FacadeOptimizationModelType, AbstractOptimizationModelType
    ],
    abstract.Retriever,
):
    def get(self, name: str) -> FacadeOptimizationModelType:
        model = self._backend_repository.get(run_id=self._run.id, name=name)
        return self._model_type(_backend=self.backend, _model=model, _run=self._run)


class Lister(
    OptimizationBaseRepository[
        FacadeOptimizationModelType, AbstractOptimizationModelType
    ],
    abstract.Lister,
):
    def list(self, name: str | None = None) -> list[FacadeOptimizationModelType]:
        models = self._backend_repository.list(run_id=self._run.id, name=name)
        return [
            self._model_type(_backend=self.backend, _model=m, _run=self._run)
            for m in models
        ]


class Tabulator(
    OptimizationBaseRepository[
        FacadeOptimizationModelType, AbstractOptimizationModelType
    ],
    abstract.Tabulator,
):
    def tabulate(self, name: str | None = None) -> pd.DataFrame:
        return self._backend_repository.tabulate(run_id=self._run.id, name=name)


class EnumerateKwargs(TypedDict, total=False):
    name: str | None


class Enumerator(
    Lister[FacadeOptimizationModelType, AbstractOptimizationModelType],
    Tabulator[FacadeOptimizationModelType, AbstractOptimizationModelType],
    abstract.Enumerator,
):
    def enumerate(
        self, table: bool = False, **kwargs: Unpack[EnumerateKwargs]
    ) -> list[FacadeOptimizationModelType] | pd.DataFrame:
        return self.tabulate(**kwargs) if table else self.list(**kwargs)



from .meta import RunMetaEntryFilter
from .model import ModelFilter
from .optimizationequation import (
    OptimizationEquationFilter,
    OptimizationEquationIndexSetAssociationFilter,
)
from .optimizationindexset import (
    OptimizationIndexSetDataFilter,
    OptimizationIndexSetFilter,
)
from .optimizationparameter import (
    OptimizationParameterFilter,
    OptimizationParameterIndexSetAssociationFilter,
)
from .optimizationscalar import OptimizationScalarFilter
from .optimizationtable import (
    OptimizationTableFilter,
    OptimizationTableIndexSetAssociationFilter,
)
from .optimizationvariable import (
    OptimizationVariableFilter,
    OptimizationVariableIndexSetAssociationFilter,
)
from .region import RegionFilter
from .run import RunFilter
from .scenario import ScenarioFilter
from .timeseries import TimeSeriesFilter
from .unit import UnitFilter
from .variable import VariableFilter



from typing import ClassVar

from ixmp4.db import Session, filters, sql

from .. import Run, Scalar, Unit


class OptimizationScalarFilter(filters.BaseFilter, metaclass=filters.FilterMeta):
    id: filters.Id | None = filters.Field(None)
    name: filters.String | None = filters.Field(None)
    run__id: filters.Integer | None = filters.Field(None, alias="run_id")
    unit__id: filters.Integer | None = filters.Field(None, alias="unit_id")

    # revert_platform: filters.Boolean = filters.Field(False)

    sqla_model: ClassVar[type] = Scalar

    def join(
        self, exc: sql.Select[tuple[Scalar]], session: Session | None = None
    ) -> sql.Select[tuple[Scalar]]:
        exc = exc.join(Run, onclause=Scalar.run__id == Run.id)
        exc = exc.join(Unit, onclause=Scalar.unit__id == Unit.id)
        return exc



# type: ignore
"""Normalize Indexset.data storage

Revision ID: 914991d09f59
Revises: 0d73f7467dab
Create Date: 2024-10-29 15:37:51.485552

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import sqlite

# Revision identifiers, used by Alembic.
revision = "914991d09f59"
down_revision = "0d73f7467dab"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "optimization_indexsetdata",
        sa.Column("indexset__id", sa.Integer(), nullable=False),
        sa.Column("value", sa.String(), nullable=False),
        sa.Column(
            "id",
            sa.Integer(),
            sa.Identity(always=False, on_null=True, start=1, increment=1),
            nullable=False,
        ),
        sa.ForeignKeyConstraint(
            ["indexset__id"],
            ["optimization_indexset.id"],
            name=op.f(
                "fk_optimization_indexsetdata_indexset__id_optimization_indexset"
            ),
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_optimization_indexsetdata")),
        sa.UniqueConstraint(
            "indexset__id",
            "value",
            name=op.f("uq_optimization_indexsetdata_indexset__id_value"),
        ),
    )
    with op.batch_alter_table("optimization_indexsetdata", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_optimization_indexsetdata_indexset__id"),
            ["indexset__id"],
            unique=False,
        )

    with op.batch_alter_table("optimization_indexset", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column(
                "_data_type",
                sa.Enum("float", "int", "str", native_enum=False),
                nullable=True,
            )
        )
        batch_op.drop_column("elements")

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("optimization_indexset", schema=None) as batch_op:
        batch_op.add_column(sa.Column("elements", sqlite.JSON(), nullable=False))
        batch_op.drop_column("_data_type")

    with op.batch_alter_table("optimization_indexsetdata", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_optimization_indexsetdata_indexset__id"))

    op.drop_table("optimization_indexsetdata")
    # ### end Alembic commands ###



from typing import ClassVar

from ixmp4.data.db.optimization.indexset.model import IndexSetData
from ixmp4.db import Session, filters, sql

from .. import IndexSet, Run


class OptimizationIndexSetFilter(filters.BaseFilter, metaclass=filters.FilterMeta):
    id: filters.Id | None = filters.Field(None)
    name: filters.String | None = filters.Field(None)
    run__id: filters.Integer | None = filters.Field(None, alias="run_id")

    sqla_model: ClassVar[type] = IndexSet

    def join(
        self, exc: sql.Select[tuple[IndexSet]], session: Session | None = None
    ) -> sql.Select[tuple[IndexSet]]:
        exc = exc.join(Run, onclause=IndexSet.run__id == Run.id)
        return exc


class OptimizationIndexSetDataFilter(filters.BaseFilter, metaclass=filters.FilterMeta):
    id: filters.Id | None = filters.Field(None)
    indexset__id: filters.Integer | None = filters.Field(None, alias="indexset_id")

    sqla_model: ClassVar[type] = IndexSetData

    def join(
        self, exc: sql.Select[tuple[IndexSetData]], session: Session | None = None
    ) -> sql.Select[tuple[IndexSetData]]:
        exc = exc.join(IndexSet, onclause=IndexSetData.indexset__id == IndexSet.id)
        return exc



from datetime import datetime
from typing import TYPE_CHECKING, ClassVar

# TODO Import this from typing when dropping Python 3.11
from typing_extensions import Unpack

from ixmp4.data.abstract import Docs as DocsModel
from ixmp4.data.abstract import IndexSet as IndexSetModel

from .base import (
    Creator,
    Deleter,
    Lister,
    OptimizationBaseModelFacade,
    Retriever,
    Tabulator,
)

if TYPE_CHECKING:
    from ixmp4.core.run import Run

    from . import InitKwargs


class IndexSet(OptimizationBaseModelFacade):
    _model: IndexSetModel
    NotFound: ClassVar = IndexSetModel.NotFound
    NotUnique: ClassVar = IndexSetModel.NotUnique

    @property
    def id(self) -> int:
        return self._model.id

    @property
    def name(self) -> str:
        return self._model.name

    @property
    def data(self) -> list[float] | list[int] | list[str]:
        return self._model.data

    def add(
        self, data: float | int | str | list[float] | list[int] | list[str]
    ) -> None:
        """Adds data to an existing IndexSet."""
        self._run.require_lock()
        self.backend.optimization.indexsets.add_data(id=self._model.id, data=data)
        self._model = self.backend.optimization.indexsets.get(
            run_id=self._model.run__id, name=self._model.name
        )

    def remove(
        self,
        data: float | int | str | list[float] | list[int] | list[str],
        remove_dependent_data: bool = True,
    ) -> None:
        """Removes data from an existing IndexSet."""
        self._run.require_lock()
        self.backend.optimization.indexsets.remove_data(
            id=self._model.id, data=data, remove_dependent_data=remove_dependent_data
        )
        self._model = self.backend.optimization.indexsets.get(
            run_id=self._model.run__id, name=self._model.name
        )

    @property
    def run_id(self) -> int:
        return self._model.run__id

    @property
    def created_at(self) -> datetime | None:
        return self._model.created_at

    @property
    def created_by(self) -> str | None:
        return self._model.created_by

    @property
    def docs(self) -> str | None:
        try:
            return self.backend.optimization.indexsets.docs.get(self.id).description
        except DocsModel.NotFound:
            return None

    @docs.setter
    def docs(self, description: str | None) -> None:
        if description is None:
            self.backend.optimization.indexsets.docs.delete(self.id)
        else:
            self.backend.optimization.indexsets.docs.set(self.id, description)

    @docs.deleter
    def docs(self) -> None:
        try:
            self.backend.optimization.indexsets.docs.delete(self.id)
        # TODO: silently failing
        except DocsModel.NotFound:
            return None

    def __str__(self) -> str:
        return f"<IndexSet {self.id} name={self.name}>"


class IndexSetRepository(
    Creator[IndexSet, IndexSetModel],
    Deleter[IndexSet, IndexSetModel],
    Retriever[IndexSet, IndexSetModel],
    Lister[IndexSet, IndexSetModel],
    Tabulator[IndexSet, IndexSetModel],
):
    def __init__(self, _run: "Run", **kwargs: Unpack["InitKwargs"]) -> None:
        super().__init__(_run=_run, **kwargs)
        self._backend_repository = self.backend.optimization.indexsets
        self._model_type = IndexSet

    def create(self, name: str) -> IndexSet:
        return super().create(name=name)



from collections.abc import Callable
from typing import Any


def autodoc(f: Callable[..., Any]) -> Callable[..., Any]:
    funcname = f""":func:`{f.__module__}.{f.__qualname__}`\n\n"""
    f.__doc__ = funcname + f.__doc__ if f.__doc__ is not None else funcname
    return f



import logging
from collections.abc import Iterable
from typing import TYPE_CHECKING, Any, cast

import pandas as pd

# TODO Import this from typing when dropping Python 3.11
from typing_extensions import Unpack

from ixmp4 import db
from ixmp4.core.exceptions import OptimizationItemUsageError
from ixmp4.data import types
from ixmp4.data.abstract import optimization as abstract
from ixmp4.data.auth.decorators import guard
from ixmp4.data.db import versions
from ixmp4.data.db.optimization.associations import (
    BaseIndexSetAssociationRepository,
    BaseIndexSetAssociationReverter,
    BaseIndexSetAssociationVersionRepository,
)

from .. import base
from .docs import OptimizationVariableDocsRepository
from .model import OptimizationVariable as Variable
from .model import (
    VariableIndexsetAssociation,
    VariableIndexsetAssociationVersion,
    VariableVersion,
)

if TYPE_CHECKING:
    from ixmp4.data.backend.db import SqlAlchemyBackend

logger = logging.getLogger(__name__)


class VariableVersionRepository(versions.VersionRepository[VariableVersion]):
    model_class = VariableVersion


class VariableIndexSetAssociationVersionRepository(
    BaseIndexSetAssociationVersionRepository[VariableIndexsetAssociationVersion]
):
    model_class = VariableIndexsetAssociationVersion
    parent_version = VariableVersion
    _item_id_column = "variable__id"


class VariableIndexSetAssociationRepository(
    BaseIndexSetAssociationRepository[
        VariableIndexsetAssociation, VariableIndexsetAssociationVersion
    ]
):
    model_class = VariableIndexsetAssociation
    versions: VariableIndexSetAssociationVersionRepository

    def __init__(self, backend: "SqlAlchemyBackend") -> None:
        super().__init__(backend)

        self.versions = VariableIndexSetAssociationVersionRepository(backend)

        from .filter import OptimizationVariableIndexSetAssociationFilter

        self.filter_class = OptimizationVariableIndexSetAssociationFilter


class VariableRepository(
    base.Creator[Variable],
    base.Deleter[Variable],
    base.Retriever[Variable],
    base.Enumerator[Variable],
    BaseIndexSetAssociationReverter[
        VariableIndexsetAssociation, VariableIndexsetAssociationVersion
    ],
    abstract.VariableRepository,
):
    model_class = Variable

    UsageError = OptimizationItemUsageError

    versions: VariableVersionRepository

    def __init__(self, *args: "SqlAlchemyBackend") -> None:
        super().__init__(*args)
        self.docs = OptimizationVariableDocsRepository(*args)

        from .filter import OptimizationVariableFilter

        self.filter_class = OptimizationVariableFilter

        self.versions = VariableVersionRepository(*args)

        self._associations = VariableIndexSetAssociationRepository(*args)

        self._item_id_column = "variable__id"

    def add(
        self,
        run_id: int,
        name: str,
        constrained_to_indexsets: list[str] | None = None,
        column_names: list[str] | None = None,
    ) -> Variable:
        variable = Variable(name=name, run__id=run_id)

        if constrained_to_indexsets:
            indexsets = {
                indexset: self.backend.optimization.indexsets.get(
                    run_id=run_id, name=indexset
                )
                for indexset in set(constrained_to_indexsets)
            }
            for i in range(len(constrained_to_indexsets)):
                _ = VariableIndexsetAssociation(
                    variable=variable,
                    indexset=indexsets[constrained_to_indexsets[i]],
                    column_name=column_names[i] if column_names else None,
                )

        variable.set_creation_info(auth_context=self.backend.auth_context)
        self.session.add(variable)

        return variable

    @guard("view")
    def get(self, run_id: int, name: str) -> Variable:
        # TODO by avoiding super().select, don't we also miss out on filters and auth?
        exc = db.select(Variable).where(
            (Variable.name == name) & (Variable.run__id == run_id)
        )
        try:
            return self.session.execute(exc).scalar_one()
        except db.NoResultFound:
            raise Variable.NotFound

    @guard("view")
    def get_by_id(self, id: int) -> Variable:
        obj = self.session.get(self.model_class, id)

        if obj is None:
            raise Variable.NotFound(id=id)

        return obj

    @guard("edit")
    def create(
        self,
        run_id: int,
        name: str,
        constrained_to_indexsets: list[str] | None = None,
        column_names: list[str] | None = None,
    ) -> Variable:
        if column_names:
            # TODO If this is removed, need to check above that constrained_to_indexsets
            #  is not None
            if constrained_to_indexsets is None:
                raise self.UsageError(
                    f"While processing Variable {name}: \n"
                    "Received `column_names` to name columns, but no "
                    "`constrained_to_indexsets` to indicate which IndexSets to use for "
                    "these columns. Please provide `constrained_to_indexsets` or "
                    "remove `column_names`!"
                )
            elif len(column_names) != len(constrained_to_indexsets):
                raise self.UsageError(
                    f"While processing Variable {name}: \n"
                    "`constrained_to_indexsets` and `column_names` not equal in "
                    "length! Please provide the same number of entries for both!"
                )
            elif len(column_names) != len(set(column_names)):
                raise self.UsageError(
                    f"While processing Variable {name}: \n"
                    "The given `column_names` are not unique!"
                )

        return super().create(
            run_id=run_id,
            name=name,
            constrained_to_indexsets=constrained_to_indexsets,
            column_names=column_names,
        )

    @guard("edit")
    def delete(self, id: int) -> None:
        # Manually ensure associations are deleted first to fix order of version updates
        associations_to_delete = self._associations.tabulate(variable_id=id)
        self._associations.bulk_delete(associations_to_delete)
        super().delete(id=id)

    @guard("view")
    def list(self, **kwargs: Unpack["base.EnumerateKwargs"]) -> Iterable[Variable]:
        return super().list(**kwargs)

    @guard("view")
    def tabulate(self, **kwargs: Unpack["base.EnumerateKwargs"]) -> pd.DataFrame:
        return super().tabulate(**kwargs)

    @guard("edit")
    def add_data(self, id: int, data: dict[str, Any] | pd.DataFrame) -> None:
        if isinstance(data, dict):
            try:
                data = pd.DataFrame.from_dict(data=data)
            except ValueError as e:
                raise Variable.DataInvalid(str(e)) from e

        if data.empty:
            return  # nothing to do

        variable = self.get_by_id(id=id)

        missing_columns = set(["levels", "marginals"]) - set(data.columns)
        if missing_columns:
            raise OptimizationItemUsageError(
                "Variable.data must include the column(s): "
                f"{', '.join(missing_columns)}!"
            )

        index_list = variable.column_names or variable.indexset_names
        existing_data = pd.DataFrame(variable.data)
        if index_list:
            data = data.set_index(index_list)
            if not existing_data.empty:
                existing_data.set_index(index_list, inplace=True)
        data = data.combine_first(existing_data)
        if index_list:
            data = data.reset_index()

        variable.data = cast(types.JsonDict, data.to_dict(orient="list"))

        self.session.commit()

    @guard("edit")
    def remove_data(
        self, id: int, data: dict[str, Any] | pd.DataFrame | None = None
    ) -> None:
        variable = self.get_by_id(id=id)

        if data is None:
            # Remove all data per default
            # TODO Is there a better way to reset .data?
            variable.data = {}
        else:
            if isinstance(data, dict):
                data = pd.DataFrame.from_dict(data=data)

            if data.empty:
                return

            index_list = variable.column_names or variable.indexset_names
            if not index_list:
                logger.warning(
                    f"Trying to remove {data.to_dict(orient='list')} from Variable '"
                    f"{variable.name}', but that is not indexed; not removing anything!"
                )
                return  # can't remove specific data from unindexed variable

            existing_data = pd.DataFrame(variable.data)
            if not existing_data.empty:
                existing_data.set_index(index_list, inplace=True)

            # This is the only kind of validation we do for removal data
            try:
                data.set_index(index_list, inplace=True)
            except KeyError as e:
                logger.error(
                    f"Data to be removed must include {index_list} as keys/columns, "
                    f"but {[name for name in data.columns]} were provided."
                )
                raise OptimizationItemUsageError(
                    "The data to be removed must specify one or more complete indices "
                    "to remove associated levels and marginals!"
                ) from e

            remaining_data = existing_data[~existing_data.index.isin(data.index)]
            if not remaining_data.index.empty:
                remaining_data.reset_index(inplace=True)

            variable.data = cast(types.JsonDict, remaining_data.to_dict(orient="list"))

        self.session.commit()

    @guard("edit")
    def revert(self, transaction__id: int, run__id: int) -> None:
        super().revert(transaction__id=transaction__id, run__id=run__id)

        # Revert VariableIndexSetAssociation
        self._revert_indexset_association(
            repo=self, transaction__id=transaction__id, run__id=run__id
        )



"""
This module contains sqlalchemy database models and repositories.
"""

from .base import BaseModel, BaseRepository
from .checkpoint import Checkpoint, CheckpointRepository
from .docs import BaseDocsRepository, docs_model
from .iamc import (  # AnnualDataPoint,; SubAnnualDataPoint,; CategoricalDataPoint,
    DataPoint,
    DataPointRepository,
    Measurand,
    MeasurandRepository,
    TimeSeries,
    TimeSeriesRepository,
    UniversalDataPoint,
    Variable,
    VariableRepository,
)
from .meta import RunMetaEntry, RunMetaEntryRepository
from .model import Model, ModelRepository
from .optimization import (
    Equation,
    EquationRepository,
    IndexSet,
    IndexSetRepository,
    Parameter,
    ParameterRepository,
    Scalar,
    ScalarRepository,
    Table,
    TableRepository,
)
from .optimization import Variable as OptimizationVariable
from .optimization import VariableRepository as OptimizationVariableRepository
from .region import Region, RegionRepository
from .run import Run, RunRepository
from .scenario import Scenario, ScenarioRepository
from .unit import Unit, UnitRepository
from .versions import TransactionRepository



from typing import ClassVar

from ixmp4 import db
from ixmp4.data import abstract
from ixmp4.data.db import versions

from .. import base


class DataPoint(base.BaseModel):
    NotFound: ClassVar = abstract.DataPoint.NotFound
    NotUnique: ClassVar = abstract.DataPoint.NotUnique
    DeletionPrevented: ClassVar = abstract.DataPoint.DeletionPrevented

    Type: ClassVar = abstract.DataPoint.Type

    __abstract__ = True

    updateable_columns = ["value"]

    time_series__id: db.MappedColumn[int] = db.Column(
        db.Integer,
        db.ForeignKey("iamc_timeseries.id"),
        nullable=False,
        index=True,
    )

    value = db.Column(db.Float)

    type = db.Column(db.String(255), nullable=False, index=True)

    step_category = db.Column(db.String(1023), index=True)
    step_year = db.Column(db.Integer, index=True)
    step_datetime = db.Column(db.DateTime, index=True)


class UniversalDataPoint(DataPoint):
    __tablename__ = "iamc_datapoint_universal"

    __table_args__ = (
        db.UniqueConstraint("time_series__id", "step_year", "step_category"),
        db.UniqueConstraint("time_series__id", "step_datetime"),
    )


def get_datapoint_model(session: db.Session) -> type[DataPoint]:
    return UniversalDataPoint


class UniversalDataPointVersion(versions.DefaultVersionModel):
    __tablename__ = "iamc_datapoint_universal_version"

    value = db.Column(db.Float)

    time_series__id: db.MappedColumn[int] = db.Column(
        db.Integer,
        nullable=False,
        index=True,
    )

    type = db.Column(db.String(255), nullable=False, index=True)

    step_category = db.Column(db.String(1023), index=True)
    step_year = db.Column(db.Integer, index=True)
    step_datetime = db.Column(db.DateTime, index=True)


version_triggers = versions.PostgresVersionTriggers(
    UniversalDataPoint.__table__, UniversalDataPointVersion.__table__
)



from collections.abc import Iterable

# TODO Use `type` when dropping Python 3.11
from typing import TypeAlias

# TODO Import this from typing when dropping Python 3.11
from typing_extensions import NotRequired, TypedDict

PrimitiveTypes: TypeAlias = bool | float | int | str
PrimitiveIterableTypes: TypeAlias = (
    Iterable[bool] | Iterable[float] | Iterable[int] | Iterable[str]
)

IntFilterAlias: TypeAlias = int | Iterable[int]
StrFilterAlias: TypeAlias = str | Iterable[str]
DefaultFilterAlias: TypeAlias = IntFilterAlias | StrFilterAlias
OptimizationFilterAlias: TypeAlias = dict[str, DefaultFilterAlias | None]

# NOTE If you want to be nitpicky, you could argue that `timeseries` have an additional
# `variable` filter, which is not clear from this Alias used for both. However,
# `variable` only adds more of the same types and we only use this for casting, so we
# are fine *for now*.
IamcFilterAlias: TypeAlias = dict[
    str,
    bool | DefaultFilterAlias | dict[str, DefaultFilterAlias] | None,
]
IamcObjectFilterAlias: TypeAlias = dict[
    str,
    DefaultFilterAlias
    | dict[
        str,
        dict[str, DefaultFilterAlias | IamcFilterAlias],
    ]
    | bool
    | None,
]


class HasIdFilter(TypedDict, total=False):
    id: int
    id__in: Iterable[int]


class HasNameFilter(TypedDict, total=False):
    name: str | None
    name__in: Iterable[str]
    name__like: str
    name__ilike: str
    name__notlike: str
    name__notilike: str


class HasHierarchyFilter(TypedDict, total=False):
    hierarchy: str | None
    hierarchy__in: Iterable[str]
    hierarchy__like: str
    hierarchy__ilike: str
    hierarchy__notlike: str
    hierarchy__notilike: str


class HasRunIdFilter(TypedDict, total=False):
    run_id: int | None
    run_id__in: Iterable[int]
    run_id__gt: int
    run_id__lt: int
    run_id__gte: int
    run_id__lte: int
    run__id: int | None
    run__id__in: Iterable[int]
    run__id__gt: int
    run__id__lt: int
    run__id__gte: int
    run__id__lte: int


class HasUnitIdFilter(TypedDict, total=False):
    unit_id: int | None
    unit_id__in: Iterable[int]
    unit_id__gt: int
    unit_id__lt: int
    unit_id__gte: int
    unit_id__lte: int
    unit__id: int | None
    unit__id__in: Iterable[int]
    unit__id__gt: int
    unit__id__lt: int
    unit__id__gte: int
    unit__id__lte: int


class HasRegionFilter(HasHierarchyFilter, HasIdFilter, HasNameFilter, total=False): ...


class HasModelFilter(HasIdFilter, HasNameFilter, total=False): ...


class HasScenarioFilter(HasIdFilter, HasNameFilter, total=False): ...


class HasUnitFilter(HasIdFilter, HasNameFilter, total=False): ...


class HasVariableFilter(HasIdFilter, HasNameFilter, total=False): ...


class HasRunFilter(HasIdFilter, total=False):
    version: int | None
    default_only: bool
    is_default: bool | None
    model: HasModelFilter | None
    scenario: HasScenarioFilter | None


class IamcScenarioFilter(TypedDict, total=False):
    region: HasRegionFilter | None
    variable: HasVariableFilter | None
    unit: HasUnitFilter | None
    run: HasRunFilter


class IamcTimeseriesFilter(HasNameFilter, total=False):
    region: HasRegionFilter | None
    variable: HasVariableFilter | None
    unit: HasUnitFilter | None
    run: HasRunFilter


class IamcUnitFilter(TypedDict, total=False):
    region: HasRegionFilter | None
    variable: HasVariableFilter | None
    run: HasRunFilter


class IamcRunFilter(TypedDict, total=False):
    region: HasRegionFilter | None
    variable: HasVariableFilter | None
    unit: HasUnitFilter | None


class IamcRegionFilter(TypedDict, total=False):
    variable: HasVariableFilter | None
    unit: HasUnitFilter | None
    run: HasRunFilter


class IamcModelFilter(TypedDict, total=False):
    region: HasRegionFilter | None
    variable: HasVariableFilter | None
    unit: HasUnitFilter | None
    run: HasRunFilter


class HasPaginationArgs(TypedDict, total=False):
    limit: int | None
    offset: int | None


class HasTransactionIdFilter(TypedDict, total=False):
    transaction_id: int | None
    transaction_id__in: Iterable[int]
    transaction_id__gt: int
    transaction_id__lt: int
    transaction_id__gte: int
    transaction_id__lte: int
    transaction__id: int | None
    transaction__id__in: Iterable[int]
    transaction__id__gt: int
    transaction__id__lt: int
    transaction__id__gte: int
    transaction__id__lte: int


class TabulateVersionsKwargs(HasPaginationArgs, total=False):
    transaction__id: NotRequired[int]


class TabulateRunMetaVersionsKwargs(TabulateVersionsKwargs):
    run__id: NotRequired[int]


class TabulateDatapointVersionsKwargs(TabulateVersionsKwargs):
    run__id: NotRequired[int]


# TODO This seems to get repetitive
class TabulateOptimizationVersionsKwargs(TabulateVersionsKwargs):
    run__id: NotRequired[int]



import re
from typing import Any, TypeVar

import pandas as pd

from ixmp4 import db
from ixmp4.conf.manager import ManagerConfig, ManagerPlatformInfo
from ixmp4.conf.user import User
from ixmp4.data.db import Model
from ixmp4.db import utils


class AuthorizationContext(object):
    def __init__(
        self, user: User, manager: ManagerConfig, platform: ManagerPlatformInfo
    ):
        self.user = user
        self.manager = manager
        self.platform = platform

    def tabulate_permissions(self) -> pd.DataFrame:
        df = self.manager.fetch_user_permissions(
            self.user, self.platform, jti=self.user.jti
        )
        if self.platform.accessibility != ManagerPlatformInfo.Accessibilty.PRIVATE:
            group_df = self.manager.fetch_group_permissions(
                self.platform.access_group, self.platform, jti=self.user.jti
            )
            df = pd.concat([df, group_df])
        df = df.dropna()

        def convert_to_regex(m: str) -> str:
            return re.escape(m).replace("\\*", ".*")

        def convert_to_like(m: str) -> str:
            return m.replace("*", "%").replace("_", "[_]")

        df["regex"] = df["model"].apply(convert_to_regex)
        df["like"] = df["model"].apply(convert_to_like)
        return df

    ApplyType = TypeVar("ApplyType", bound=db.sql.Select[tuple[Any]])

    def apply(self, access_type: str, exc: ApplyType) -> ApplyType:
        if self.is_managed:
            return exc
        if self.user.is_superuser:
            return exc

        if utils.is_joined(exc, Model):
            perms = self.tabulate_permissions()
            if perms.empty:
                return exc.where(db.false())
            if access_type == "edit":
                perms = perms.where(perms["access_type"] == "EDIT").dropna()
            # `*` is used as wildcard in permission logic, replaced by sql-wildcard `%`
            conditions = [Model.name.like(p["like"]) for i, p in perms.iterrows()]
            exc = exc.where(db.or_(*conditions))
        return exc

    def check_access(self, access_type: str, model_name: str) -> bool:
        if self.is_managed:
            return True
        if self.user.is_superuser:
            return True

        perms = self.tabulate_permissions()
        if perms.empty:
            return False
        if access_type == "edit":
            perms = perms.where(perms["access_type"] == "EDIT").dropna()

        regex = "^" + "|".join(perms["regex"]) + "$"
        match = re.match(regex, model_name)
        return match is not None

    @property
    def is_accessible(self) -> bool:
        if self.user.is_superuser:
            return True
        if self.is_managed:
            return True
        elif self.platform.accessibility == ManagerPlatformInfo.Accessibilty.PUBLIC:
            return True
        elif self.platform.accessibility == ManagerPlatformInfo.Accessibilty.GATED:
            return self.user.is_authenticated

        return self.platform.access_group in self.user.groups

    @property
    def is_viewable(self) -> bool:
        if self.user.is_superuser:
            return True
        if self.is_managed:
            return True

        df = self.tabulate_permissions()
        return not df[
            (df["access_type"] == "EDIT") | (df["access_type"] == "VIEW")
        ].empty

    @property
    def is_editable(self) -> bool:
        if self.user.is_superuser:
            return True
        if self.is_managed:
            return True

        df = self.tabulate_permissions()
        return not df[df["access_type"] == "EDIT"].empty

    @property
    def is_managed(self) -> bool:
        if self.user.is_superuser:
            return True
        return self.platform.management_group in self.user.groups



from ixmp4.data.db import filters as base
from ixmp4.data.db.run import Run
from ixmp4.db import Session, filters, sql, utils

from .model import Parameter, ParameterIndexsetAssociation


class RunFilter(base.RunFilter, metaclass=filters.FilterMeta):
    def join(
        self, exc: sql.Select[tuple[Parameter]], session: Session | None = None
    ) -> sql.Select[tuple[Parameter]]:
        if not utils.is_joined(exc, Run):
            exc = exc.join(Run, onclause=Parameter.run__id == Run.id)
        return exc


class OptimizationParameterFilter(
    base.OptimizationParameterFilter, metaclass=filters.FilterMeta
):
    def join(
        self, exc: sql.Select[tuple[Parameter]], session: Session | None = None
    ) -> sql.Select[tuple[Parameter]]:
        return exc


class OptimizationParameterIndexSetAssociationFilter(
    base.OptimizationParameterIndexSetAssociationFilter, metaclass=filters.FilterMeta
):
    def join(
        self,
        exc: sql.Select[tuple[ParameterIndexsetAssociation]],
        session: Session | None = None,
    ) -> sql.Select[tuple[ParameterIndexsetAssociation]]:
        return exc



from typing import Any, ClassVar, TypedDict

import pandas as pd

# TODO Import this from typing when dropping Python 3.11
from typing_extensions import Unpack

from ixmp4.core.exceptions import IxmpError
from ixmp4.data import types
from ixmp4.data.abstract.annotations import HasIdFilter, HasNameFilter, HasRunIdFilter
from ixmp4.data.db.versions.repository import VersionRepository
from ixmp4.db.filters import BaseFilter

from .. import mixins
from ..base import BaseModel as RootBaseModel
from ..base import (
    BulkDeleter,
    BulkUpserter,
    Creator,
    Deleter,
    Enumerator,
    Lister,
    ModelType,
    Retriever,
    Selecter,
    Tabulator,
)


class BaseModel(RootBaseModel, mixins.HasCreationInfo):
    # NOTE: only subclasses storing data actually define this!
    DataInvalid: ClassVar[type[IxmpError]]

    __abstract__ = True

    name: types.Name


class EnumerateKwargs(HasIdFilter, HasNameFilter, HasRunIdFilter, total=False):
    _filter: BaseFilter


class RevertKwargs(TypedDict, total=False):
    run__id: int


class Reverter(BulkDeleter[ModelType], BulkUpserter[ModelType]):
    versions: VersionRepository[Any]

    # NOTE These are stored as a baseline
    _columns_to_drop_for_update = {
        "transaction_id",
        "end_transaction_id",
        "operation_type",
    }
    _columns_to_drop_for_insert = _columns_to_drop_for_update | {"id"}

    def revert(
        self,
        transaction__id: int,
        correct_versions: pd.DataFrame | None = None,
        **kwargs: Unpack[RevertKwargs],
    ) -> None:
        current_versions = self.tabulate(**kwargs)
        if correct_versions is None:
            correct_versions = self.versions.tabulate(
                transaction__id=transaction__id, **kwargs
            )

        # TODO Add a check to exit early if current == correct

        versions_to_delete = current_versions[
            ~current_versions["id"].isin(correct_versions["id"])
        ]

        if not versions_to_delete.empty:
            self.bulk_delete(versions_to_delete)

        # NOTE The following manual data split could be avoided if `tabulate_existing`
        # returned the correct values
        # This would allow us to simply `bulk_upsert` the data, though that queries the
        # DB again (for `tabulate_existing`, the data of which we already have here)

        versions_to_insert = correct_versions[
            ~correct_versions["id"].isin(current_versions["id"])
        ].drop(columns=self._columns_to_drop_for_insert)

        if not versions_to_insert.empty:
            with self.backend.event_handler.pause():
                self.bulk_insert(versions_to_insert)

        # Limit updates to those differing from their existing versions
        ids_to_update = set(
            self.versions.tabulate(
                transaction__id=transaction__id, valid="after_transaction", **kwargs
            )["id"]
        )
        correct_versions = correct_versions[correct_versions["id"].isin(ids_to_update)]

        versions_to_update = correct_versions[
            correct_versions["id"].isin(current_versions["id"])
        ].drop(columns=self._columns_to_drop_for_update)

        if not versions_to_update.empty:
            self.bulk_update(versions_to_update)

        self.session.commit()



from collections.abc import Callable
from functools import wraps
from typing import TYPE_CHECKING, Any, Concatenate, ParamSpec, TypeVar

from ixmp4.core.exceptions import Forbidden, ProgrammingError

if TYPE_CHECKING:
    from ..db.base import BaseModel, BaseRepository

P = ParamSpec("P")

ReturnT = TypeVar("ReturnT")


def guard(
    access: str,
) -> Callable[
    [Callable[Concatenate[Any, P], ReturnT]], Callable[Concatenate[Any, P], ReturnT]
]:
    if access not in ["edit", "manage", "view"]:
        raise ProgrammingError("Guard access must be 'edit', 'manage' or 'view'.")

    def decorator(
        func: Callable[Concatenate[Any, P], ReturnT],
    ) -> Callable[Concatenate[Any, P], ReturnT]:
        @wraps(func)
        def guarded_func(
            self: "BaseRepository[BaseModel]", /, *args: P.args, **kwargs: P.kwargs
        ) -> ReturnT:
            if self.backend.auth_context is not None:
                if access == "view" and self.backend.auth_context.is_viewable:
                    return func(self, *args, **kwargs)
                elif access == "edit" and self.backend.auth_context.is_editable:
                    return func(self, *args, **kwargs)
                elif access == "manage" and self.backend.auth_context.is_managed:
                    return func(self, *args, **kwargs)

                raise Forbidden(
                    f"Function '{func.__name__}': '{access}' access denied due to "
                    "insufficient permissions."
                )
            else:
                # ignoring authorization since no auth context is set
                return func(self, *args, **kwargs)

        return guarded_func

    return decorator



import logging
from collections.abc import Iterable
from typing import TYPE_CHECKING, Any, cast

import pandas as pd

# TODO Import this from typing when dropping Python 3.11
from typing_extensions import Unpack

from ixmp4 import db
from ixmp4.core.exceptions import OptimizationItemUsageError
from ixmp4.data import types
from ixmp4.data.abstract import optimization as abstract
from ixmp4.data.auth.decorators import guard
from ixmp4.data.db import versions
from ixmp4.data.db.optimization.associations import (
    BaseIndexSetAssociationRepository,
    BaseIndexSetAssociationReverter,
    BaseIndexSetAssociationVersionRepository,
)

from .. import base
from .docs import TableDocsRepository
from .model import (
    Table,
    TableIndexsetAssociation,
    TableIndexsetAssociationVersion,
    TableVersion,
)

if TYPE_CHECKING:
    from ixmp4.data.backend.db import SqlAlchemyBackend

logger = logging.getLogger(__name__)


class TableVersionRepository(versions.VersionRepository[TableVersion]):
    model_class = TableVersion


class TableIndexSetAssociationVersionRepository(
    BaseIndexSetAssociationVersionRepository[TableIndexsetAssociationVersion]
):
    model_class = TableIndexsetAssociationVersion
    parent_version = TableVersion
    _item_id_column = "table__id"


class TableIndexSetAssociationRepository(
    BaseIndexSetAssociationRepository[
        TableIndexsetAssociation, TableIndexsetAssociationVersion
    ]
):
    model_class = TableIndexsetAssociation
    versions: TableIndexSetAssociationVersionRepository

    def __init__(self, backend: "SqlAlchemyBackend") -> None:
        super().__init__(backend)

        self.versions = TableIndexSetAssociationVersionRepository(backend)

        from .filter import OptimizationTableIndexSetAssociationFilter

        self.filter_class = OptimizationTableIndexSetAssociationFilter


class TableRepository(
    base.Creator[Table],
    base.Deleter[Table],
    base.Retriever[Table],
    base.Enumerator[Table],
    BaseIndexSetAssociationReverter[
        TableIndexsetAssociation, TableIndexsetAssociationVersion
    ],
    abstract.TableRepository,
):
    model_class = Table

    UsageError = OptimizationItemUsageError

    versions: TableVersionRepository

    def __init__(self, *args: "SqlAlchemyBackend") -> None:
        super().__init__(*args)
        self.docs = TableDocsRepository(*args)

        from .filter import OptimizationTableFilter

        self.filter_class = OptimizationTableFilter

        self.versions = TableVersionRepository(*args)

        self._associations = TableIndexSetAssociationRepository(*args)

        self._item_id_column = "table__id"

    def add(
        self,
        run_id: int,
        name: str,
        constrained_to_indexsets: list[str],
        column_names: list[str] | None = None,
    ) -> Table:
        table = Table(name=name, run__id=run_id)

        # TODO fix this for other items and add tests for all, too
        indexsets = {
            indexset: self.backend.optimization.indexsets.get(
                run_id=run_id, name=indexset
            )
            for indexset in set(constrained_to_indexsets)
        }
        for i in range(len(constrained_to_indexsets)):
            _ = TableIndexsetAssociation(
                table=table,
                indexset=indexsets[constrained_to_indexsets[i]],
                column_name=column_names[i] if column_names else None,
            )

        table.set_creation_info(auth_context=self.backend.auth_context)
        self.session.add(table)

        return table

    @guard("view")
    def get(self, run_id: int, name: str) -> Table:
        exc = db.select(Table).where((Table.name == name) & (Table.run__id == run_id))
        try:
            return self.session.execute(exc).scalar_one()
        except db.NoResultFound:
            raise Table.NotFound

    @guard("view")
    def get_by_id(self, id: int) -> Table:
        obj = self.session.get(self.model_class, id)

        if obj is None:
            raise Table.NotFound(id=id)

        return obj

    @guard("edit")
    def create(
        self,
        run_id: int,
        name: str,
        constrained_to_indexsets: list[str],
        column_names: list[str] | None = None,
    ) -> Table:
        if column_names and len(column_names) != len(constrained_to_indexsets):
            raise self.UsageError(
                f"While processing Table {name}: \n"
                "`constrained_to_indexsets` and `column_names` not equal in length! "
                "Please provide the same number of entries for both!"
            )

        if column_names and len(column_names) != len(set(column_names)):
            raise self.UsageError(
                f"While processing Table {name}: \n"
                "The given `column_names` are not unique!"
            )

        return super().create(
            run_id=run_id,
            name=name,
            constrained_to_indexsets=constrained_to_indexsets,
            column_names=column_names,
        )

    @guard("edit")
    def delete(self, id: int) -> None:
        # Manually ensure associations are deleted first to fix order of version updates
        associations_to_delete = self._associations.tabulate(table_id=id)
        self._associations.bulk_delete(associations_to_delete)
        super().delete(id=id)

    @guard("view")
    def list(self, **kwargs: Unpack["base.EnumerateKwargs"]) -> Iterable[Table]:
        return super().list(**kwargs)

    @guard("view")
    def tabulate(self, **kwargs: Unpack["base.EnumerateKwargs"]) -> pd.DataFrame:
        return super().tabulate(**kwargs)

    @guard("edit")
    def add_data(self, id: int, data: dict[str, Any] | pd.DataFrame) -> None:
        if isinstance(data, dict):
            data = pd.DataFrame.from_dict(data=data)

        if data.empty:
            return  # nothing to do

        table = self.get_by_id(id=id)

        table.data = cast(
            types.JsonDict,
            pd.concat([pd.DataFrame.from_dict(table.data), data]).to_dict(
                orient="list"
            ),
        )

        # Move pending changes to the DB transaction buffer
        self.session.commit()

    @guard("edit")
    def remove_data(self, id: int, data: dict[str, Any] | pd.DataFrame) -> None:
        if isinstance(data, dict):
            data = pd.DataFrame.from_dict(data=data)

        if data.empty:
            return

        table = self.get_by_id(id=id)
        index_list = table.column_names or table.indexset_names
        existing_data = pd.DataFrame(table.data)
        if not existing_data.empty:
            existing_data.set_index(index_list, inplace=True)

        # This is the only kind of validation we do for removal data
        try:
            data.set_index(index_list, inplace=True)
        except KeyError as e:
            logger.error(
                f"Data to be removed must include {index_list} as keys/columns, but "
                f"{[name for name in data.columns]} were provided."
            )
            raise OptimizationItemUsageError(
                "The data to be removed must specify one or more complete indices!"
            ) from e

        remaining_data = existing_data[~existing_data.index.isin(data.index)]
        if not remaining_data.index.empty:
            remaining_data.reset_index(inplace=True)

        table.data = cast(types.JsonDict, remaining_data.to_dict(orient="list"))

        self.session.commit()

    @guard("edit")
    def revert(self, transaction__id: int, run__id: int) -> None:
        super().revert(transaction__id=transaction__id, run__id=run__id)

        # Revert TableIndexSetAssociation
        self._revert_indexset_association(
            repo=self, transaction__id=transaction__id, run__id=run__id
        )



from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, Generic, Protocol, TypeVar

if TYPE_CHECKING:
    from ixmp4.data.db.timeseries import CreateKwargs

import pandas as pd

# TODO Import this from typing when dropping Python 3.11
from typing_extensions import Unpack

from .. import base
from ..annotations import (
    HasIdFilter,
    HasRegionFilter,
    HasRunFilter,
    HasUnitFilter,
    HasVariableFilter,
)


class TimeSeries(base.BaseModel, Protocol):
    """Time series data model."""

    run__id: int
    "Unique run id."
    parameters: Mapping[str, Any]
    "A set of parameter values for the time series."

    def __str__(self) -> str:
        return f"<TimeSeries {self.id} parameters={self.parameters}>"


ModelType = TypeVar("ModelType", bound=TimeSeries)


class EnumerateKwargs(HasIdFilter, total=False):
    region: HasRegionFilter
    unit: HasUnitFilter
    variable: HasVariableFilter
    run: HasRunFilter


class TimeSeriesRepository(
    base.Creator,
    base.Retriever,
    base.Enumerator,
    base.BulkUpserter,
    Protocol,
    Generic[ModelType],
):
    def create(self, **kwargs: Unpack["CreateKwargs"]) -> ModelType:
        """Retrieves a time series.

        Parameters
        ----------
        run__id : int
            Unique run id.
        parameters : Mapping[str, Any]
            A set of parameter values for the time series.

        Raises
        ------
        :class:`ixmp4.data.abstract.iamc.timeseries.TimeSeries.NotUnique`.

        Returns
        -------
        :class:`ixmp4.data.abstract.iamc.timeseries.TimeSeries`:
            The retrieved time series.
        """
        ...

    # NOTE this seems unused, so I'm guessing at the parameters type
    def get(self, run_id: int, parameters: Mapping[str, Any]) -> ModelType:
        """Retrieves a time series.

        Parameters
        ----------
        run_id : int
            Unique run id.
        parameters : Mapping[str, Any]
            A set of parameter values for the time series.

        Raises
        ------
        :class:`ixmp4.data.abstract.TimeSeries.NotFound`.
            If the time series with `run_id` and `parameters` does not exist.

        Returns
        -------
        :class:`ixmp4.data.base.TimeSeries`:
            The retrieved time series.
        """
        ...

    def get_by_id(self, id: int) -> ModelType:
        """Retrieves a time series by it's id.

        Parameters
        ----------
        id : int
            Unique integer id.

        Raises
        ------
        :class:`ixmp4.data.abstract.TimeSeries.NotFound`.
            If the time series with `id` does not exist.

        Returns
        -------
        :class:`ixmp4.data.base.TimeSeries`:
            The retrieved time series.
        """
        ...

    def get_or_create(self, run_id: int, parameters: Mapping[str, Any]) -> ModelType:
        """Tries to retrieve a time series and creates it if it was not found.

        Parameters
        ----------
        run_id : int
            Unique run id.
        parameters : Mapping[str, Any]
            A set of parameter values for the time series.

        Returns
        -------
        :class:`ixmp4.data.base.TimeSeries`:
            The retrieved or created time series.
        """
        try:
            return self.get(run_id, parameters)
        except TimeSeries.NotFound:
            return self.create(run__id=run_id, parameters=parameters)

    def list(self, **kwargs: Unpack[EnumerateKwargs]) -> list[ModelType]:
        r"""Lists time series by specified criteria.

        Parameters
        ----------
            \*\*kwargs: any
            Filter parameters as specified in
            :class:`ixmp4.data.db.iamc.timeseries.filter.TimeSeriesFilter`.

        Returns
        -------
        Iterable[:class:`ixmp4.data.base.TimeSeries`]:
            List of time series.
        """
        ...

    def tabulate(
        self, *, join_parameters: bool | None = False, **kwargs: Unpack[EnumerateKwargs]
    ) -> pd.DataFrame:
        r"""Tabulate time series by specified criteria.

        Parameters
        ----------
        join_parameters : bool | None
            If set to `True` the resulting data frame will include
            parameter columns as values instead of foreign key id's.
         \*\*kwargs: any
            Filter parameters as specified in
            :class:`ixmp4.data.db.iamc.timeseries.filter.TimeSeriesFilter`.

        Returns
        -------
        :class:`pandas.DataFrame`:
            A data frame with the columns:
                - id
                - run__id
                - ... parameter id columns
                  Or:
                - ... parameter value columns
        """
        ...

    def bulk_upsert(
        self, df: pd.DataFrame, create_related: bool | None = False
    ) -> None:
        """Looks which time series in the supplied data frame already exist, and
        inserts new ones.

        Parameters
        ----------
        df : :class:`pandas.DataFrame`
            A data frame with the columns:
                - id
                - run__id

                if `create_related` = False:
                    - ... parameter id columns

                else:
                    - ... parameter value columns (i.e `region` column with value
                      "World" instead of `region__id` column with value "1".)

        create_related : bool
            Creates related database entries with value data supplied in `df`.
        """
        ...



from datetime import datetime
from typing import TYPE_CHECKING, ClassVar

# TODO Import this from typing when dropping Python 3.11
from typing_extensions import Unpack

from ixmp4.core.unit import Unit
from ixmp4.data.abstract import Docs as DocsModel
from ixmp4.data.abstract import Scalar as ScalarModel
from ixmp4.data.abstract import Unit as UnitModel

from .base import Deleter, Lister, OptimizationBaseModelFacade, Retriever, Tabulator

if TYPE_CHECKING:
    from ixmp4.core.run import Run

    from . import InitKwargs


class Scalar(OptimizationBaseModelFacade):
    _model: ScalarModel
    NotFound: ClassVar = ScalarModel.NotFound
    NotUnique: ClassVar = ScalarModel.NotUnique

    @property
    def id(self) -> int:
        return self._model.id

    @property
    def name(self) -> str:
        return self._model.name

    @property
    def value(self) -> float:
        """Associated value."""
        return self._model.value

    @value.setter
    def value(self, value: float) -> None:
        self._run.require_lock()
        self._model.value = value
        self.backend.optimization.scalars.update(
            id=self._model.id, value=self._model.value
        )

    @property
    def unit(self) -> UnitModel:
        """Associated unit."""
        return self._model.unit

    @unit.setter
    def unit(self, value: str | Unit) -> None:
        self._run.require_lock()
        if isinstance(value, str):
            unit_model = self.backend.units.get(value)
            value = Unit(_backend=self.backend, _model=unit_model)
        self._model = self.backend.optimization.scalars.update(
            id=self._model.id, unit_id=value.id
        )

    @property
    def run_id(self) -> int:
        return self._model.run__id

    @property
    def created_at(self) -> datetime | None:
        return self._model.created_at

    @property
    def created_by(self) -> str | None:
        return self._model.created_by

    @property
    def docs(self) -> str | None:
        try:
            return self.backend.optimization.scalars.docs.get(self.id).description
        except DocsModel.NotFound:
            return None

    @docs.setter
    def docs(self, description: str | None) -> None:
        if description is None:
            self.backend.optimization.scalars.docs.delete(self.id)
        else:
            self.backend.optimization.scalars.docs.set(self.id, description)

    @docs.deleter
    def docs(self) -> None:
        try:
            self.backend.optimization.scalars.docs.delete(self.id)
        # TODO: silently failing
        except DocsModel.NotFound:
            return None

    def __str__(self) -> str:
        return f"<Scalar {self.id} name={self.name}>"


class ScalarRepository(
    Deleter[Scalar, ScalarModel],
    Retriever[Scalar, ScalarModel],
    Lister[Scalar, ScalarModel],
    Tabulator[Scalar, ScalarModel],
):
    def __init__(self, _run: "Run", **kwargs: Unpack["InitKwargs"]) -> None:
        super().__init__(_run=_run, **kwargs)
        self._backend_repository = self.backend.optimization.scalars
        self._model_type = Scalar

    def create(self, name: str, value: float, unit: str | Unit | None = None) -> Scalar:
        self._run.require_lock()
        if isinstance(unit, Unit):
            unit_name = unit.name
        elif isinstance(unit, str):
            unit_name = unit
        else:
            # TODO: provide logging information about None-units being converted
            # if unit is None, assume that this is a dimensionless scalar (unit = "")
            dimensionless_unit = self.backend.units.get_or_create(name="")
            unit_name = dimensionless_unit.name

        try:
            model = self.backend.optimization.scalars.create(
                name=name, value=value, unit_name=unit_name, run_id=self._run.id
            )
        except Scalar.NotUnique as e:
            raise Scalar.NotUnique(
                message=f"Scalar '{name}' already exists! Did you mean to call "
                "run.optimization.scalars.update()?"
            ) from e
        return Scalar(_backend=self.backend, _model=model, _run=self._run)



from collections.abc import Iterable
from contextlib import suppress
from functools import reduce
from typing import TYPE_CHECKING, Any, TypeVar

from sqlalchemy import inspect, sql
from sqlalchemy.sql import ColumnCollection, ColumnElement
from sqlalchemy.sql.base import ReadOnlyColumnCollection

from ixmp4.core.exceptions import ProgrammingError

if TYPE_CHECKING:
    from ixmp4.data.db.base import BaseModel

JoinType = TypeVar("JoinType", bound=sql.Select[tuple["BaseModel", ...]])


# This should not need to be Any, I think, but if I put "BaseModel" instead, various
# things like RunMetaEntry are not recognized -> this sounds like covariance again,
# but covariant typevars are not allowed as type hints.
def is_joined(exc: sql.Select[tuple[Any, ...]], model: type["BaseModel"]) -> bool:
    """Returns `True` if `model` has been joined in `exc`."""
    for visitor in sql.visitors.iterate(exc):
        # Checking for `.join(Child)` clauses
        if visitor.__visit_name__ == "table":
            # Visitor might be of ColumnCollection or so,
            # which cannot be compared to model
            with suppress(TypeError):
                if model == visitor.entity_namespace:  # type: ignore[attr-defined]
                    return True
    return False


def get_columns(
    model_class: type["BaseModel"],
) -> ReadOnlyColumnCollection[str, ColumnElement[Any]]:
    mapper = inspect(model_class)
    if mapper is not None:
        return mapper.selectable.columns
    else:
        raise ProgrammingError(f"Model class `{model_class.__name__}` is not mapped.")


def _maybe_add_pk_column_to_collection(
    columns: ColumnCollection[str, ColumnElement[int]], column: ColumnElement[Any]
) -> ColumnCollection[str, ColumnElement[int]]:
    if column.primary_key:
        columns.add(column)

    return columns


def get_pk_columns(
    model_class: type["BaseModel"],
) -> ReadOnlyColumnCollection[str, ColumnElement[int]]:
    columns: ColumnCollection[str, ColumnElement[int]] = reduce(
        _maybe_add_pk_column_to_collection, get_columns(model_class), ColumnCollection()
    )

    return columns.as_readonly()


def _maybe_add_fk_column_to_collection(
    columns: ColumnCollection[str, ColumnElement[int]], column: ColumnElement[Any]
) -> ColumnCollection[str, ColumnElement[int]]:
    if len(column.foreign_keys) > 0:
        columns.add(column)

    return columns


def get_foreign_columns(
    model_class: type["BaseModel"],
) -> ReadOnlyColumnCollection[str, ColumnElement[int]]:
    columns: ColumnCollection[str, ColumnElement[int]] = reduce(
        _maybe_add_fk_column_to_collection, get_columns(model_class), ColumnCollection()
    )

    return columns.as_readonly()


def create_id_map_subquery(
    old_exc: sql.Subquery, new_exc: sql.Subquery
) -> sql.Subquery:
    return (
        sql.select(old_exc.c.id.label("old_id"), new_exc.c.id.label("new_id"))
        .join(new_exc, old_exc.c.name == new_exc.c.name)
        .subquery()
    )


def collect_columns_to_select(
    columns: ColumnCollection[str, ColumnElement[Any]], exclude: Iterable[str]
) -> ColumnCollection[str, ColumnElement[Any]]:
    # NOTE we only get ReadOnlyCollections, so can't just remove() items
    columns_to_collect: ColumnCollection[str, ColumnElement[Any]] = ColumnCollection()

    for name, column in columns.items():
        if name not in exclude:
            columns_to_collect.add(column=column, key=name)

    return columns_to_collect


def where_matches_kwargs(
    exc: sql.Select[Any], model_class: type["BaseModel"], **kwargs: Any
) -> sql.Select[Any]:
    for key in kwargs:
        columns = get_columns(model_class)
        if key in columns:
            exc = exc.where(columns[key] == kwargs[key])
    return exc



import enum
from collections.abc import Iterable
from datetime import datetime
from typing import Protocol

import pandas as pd

# TODO Import this from typing when dropping Python 3.11
from typing_extensions import TypedDict, Unpack

from .. import base


class DataPoint(base.BaseModel, Protocol):
    """Data point data model."""

    time_series__id: int
    "Foreign unique integer id of the associated time series."
    value: float
    "Value of the data point."
    type: str
    "Type of data point either `ANNUAL`, `CATEGORICAL` or `DATETIME`."

    step_year: int | None
    "An integer time step required by data points of type `ANNUAL` or `CATEGORICAL`."
    step_category: str | None
    "A string category required by data points of type `CATEGORICAL`."
    step_datetime: datetime | None
    "A datetime object required by data points of type `DATETIME`."

    class Type(str, enum.Enum):
        BASE = "BASE"
        ANNUAL = "ANNUAL"
        CATEGORICAL = "CATEGORICAL"
        DATETIME = "DATETIME"

    def __str__(self) -> str:
        return f"<Datapoint {self.id} type={self.type}>"


class EnumerateKwargs(TypedDict, total=False):
    step_year: int | None
    step_year__in: Iterable[int]
    step_year__gt: int
    step_year__lt: int
    step_year__gte: int
    step_year__lte: int
    year: int | None
    year__in: Iterable[int]
    year__gt: int
    year__lt: int
    year__gte: int
    year__lte: int
    time_series_id: int | None
    time_series_id__in: Iterable[int]
    time_series_id__gt: int
    time_series_id__lt: int
    time_series_id__gte: int
    time_series_id__lte: int
    time_series__id: int | None
    time_series__id__in: Iterable[int]
    time_series__id__gt: int
    time_series__id__lt: int
    time_series__id__gte: int
    time_series__id__lte: int
    region: dict[str, str | Iterable[str]] | None
    unit: dict[str, str | Iterable[str]] | None
    variable: dict[str, str | Iterable[str]] | None
    model: dict[str, str | Iterable[str]] | None
    scenario: dict[str, str | Iterable[str]] | None
    run: dict[str, bool | int | Iterable[int]]


class DataPointRepository(
    base.Enumerator,
    base.BulkUpserter,
    base.BulkDeleter,
    Protocol,
):
    def list(
        self,
        *,
        join_parameters: bool | None = False,
        join_runs: bool = False,
        join_run_id: bool = False,
        **kwargs: Unpack[EnumerateKwargs],
    ) -> list[DataPoint]:
        """Lists data points by specified criteria.
        This method incurrs mentionable overhead compared to :meth:`tabulate`.

        Parameters
        ----------
        join_parameters : bool | None
            If set to `True` the resulting list will include parameter columns
            from the associated :class:`ixmp4.data.abstract.iamc.timeseries.TimeSeries`.
        join_runs : bool
            If set to `True` the resulting list will include model & scenario name
            and version id of the associated Run.
        join_run_id : bool
            If set to `True` the resulting list will include the id of the associated
            Run as `run__id`.
        kwargs : Any
            Additional key word arguments. Any left over kwargs will be used as filters.

        Returns
        -------
        Iterable[:class:`ixmp4.data.abstract.iamc.datapoint.DataPoint`]:
            List of data points.
        """
        ...

    def tabulate(
        self,
        *,
        join_parameters: bool | None = False,
        join_runs: bool = False,
        join_run_id: bool = False,
        **kwargs: Unpack[EnumerateKwargs],
    ) -> pd.DataFrame:
        """Tabulates data points by specified criteria.

        Parameters
        ----------
        join_parameters : bool | None
            If set to `True` the resulting data frame will include parameter columns
            from the associated :class:`ixmp4.data.abstract.iamc.timeseries.TimeSeries`.
        join_runs : bool
            If set to `True` the resulting data frame will include model & scenario name
            and version id of the associated Run.
        join_run_id : bool
            If set to `True` the resulting data frame will include the id of the
            associated Run as `run__id`.
        kwargs : Any
            Additional key word arguments. Any left over kwargs will be used as filters.

        Returns
        -------
        :class:`pandas.DataFrame`:
            A data frame with the columns:
                - id
                - time_series__id
                - value
                - type
                - step_year
                    if it contains data points of type `ANNUAL` or `CATEGORICAL`
                - step_category
                    if it contains data points of type `CATEGORICAL`
                - step_datetime
                    if it contains data points of type `DATETIME`
                - run__id (if `join_run_id` is set to `True`)
                - ... misc parameter columns if `join_parameters` is set to `True`
        """
        ...

    def bulk_upsert(self, df: pd.DataFrame) -> None:
        """Looks which data points in the supplied data frame already exists,
        updates those that have changed and inserts new ones.

        Parameters
        ----------
        df : :class:`pandas.DataFrame`
            A data frame with the columns:
                - time_series__id
                - value
                - type
                - step_year
                    if it contains data points of type `ANNUAL` or `CATEGORICAL`
                - step_category
                    if it contains data points of type `CATEGORICAL`
                - step_datetime
                    if it contains data points of type `DATETIME`
        """
        ...

    def bulk_delete(self, df: pd.DataFrame) -> None:
        """Deletes data points which match criteria in the supplied data frame.

        Parameters
        ----------
        df : :class:`pandas.DataFrame`
            A data frame with the columns:
                - time_series__id
                - type
                - step_year
                    if it contains data points of type `ANNUAL` or `CATEGORICAL`
                - step_category
                    if it contains data points of type `CATEGORICAL`
                - step_datetime
                    if it contains data points of type `DATETIME`
        """
        ...



from collections.abc import Iterable, Mapping
from datetime import datetime
from typing import TYPE_CHECKING, Any, ClassVar, cast

import pandas as pd

# TODO Import this from typing when dropping support for Python 3.11
from typing_extensions import Unpack

from ixmp4.data import abstract

from .. import base
from ..docs import Docs, DocsRepository
from ..unit import Unit

if TYPE_CHECKING:
    from ixmp4.data.backend.api import RestBackend


class Scalar(base.BaseModel):
    NotFound: ClassVar = abstract.Scalar.NotFound
    NotUnique: ClassVar = abstract.Scalar.NotUnique
    DeletionPrevented: ClassVar = abstract.Scalar.DeletionPrevented

    id: int
    name: str
    value: float
    unit: Unit
    unit__id: int
    run__id: int

    created_at: datetime | None
    created_by: str | None


class ScalarDocsRepository(DocsRepository):
    model_class = Docs
    prefix = "docs/optimization/scalars/"


class ScalarRepository(
    base.Creator[Scalar],
    base.Deleter[Scalar],
    base.Retriever[Scalar],
    base.Enumerator[Scalar],
    abstract.ScalarRepository,
):
    model_class = Scalar
    prefix = "optimization/scalars/"

    def __init__(self, *args: Unpack[tuple["RestBackend"]]) -> None:
        super().__init__(*args)
        self.docs = ScalarDocsRepository(self.backend)

    def create(self, name: str, value: float, unit_name: str, run_id: int) -> Scalar:
        return super().create(
            name=name, value=value, unit_name=unit_name, run_id=run_id
        )

    def delete(self, id: int) -> None:
        super().delete(id=id)

    def update(
        self, id: int, value: float | None = None, unit_id: int | None = None
    ) -> Scalar:
        # we can assume this type on update endpoints
        res: Mapping[str, Any] = self._request(
            "PATCH",
            self.prefix + str(id) + "/",
            json={
                "value": value,
                "unit_id": unit_id,
            },
        )  # type: ignore[assignment]
        return self.model_class(**res)

    def get(self, run_id: int, name: str) -> Scalar:
        return super().get(run_id=run_id, name=name)

    def get_by_id(self, id: int) -> Scalar:
        res = self._get_by_id(id)
        return Scalar(**res)

    def list(
        self, **kwargs: Unpack["abstract.optimization.scalar.EnumerateKwargs"]
    ) -> Iterable[Scalar]:
        json = cast(abstract.annotations.OptimizationFilterAlias, kwargs)
        return super()._list(json=json)

    def tabulate(
        self, **kwargs: Unpack["abstract.optimization.scalar.EnumerateKwargs"]
    ) -> pd.DataFrame:
        json = cast(abstract.annotations.OptimizationFilterAlias, kwargs)
        return super()._tabulate(json=json)

    def enumerate(
        self, **kwargs: Unpack["abstract.optimization.scalar.EnumerateKwargs"]
    ) -> Iterable[Scalar] | pd.DataFrame:
        return super().enumerate(**kwargs)



from __future__ import annotations

import logging
import sqlite3
from collections.abc import Callable, Iterable, Iterator, Mapping
from datetime import datetime
from typing import TYPE_CHECKING, Any, ClassVar, Generic, Protocol, TypeVar, cast

import numpy as np
import pandas as pd
from sqlalchemy import TextClause, event, text
from sqlalchemy.engine import Engine
from sqlalchemy.engine.interfaces import Dialect
from sqlalchemy.exc import IntegrityError, NoResultFound
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.orm import Bundle, DeclarativeBase
from sqlalchemy.orm.session import Session
from sqlalchemy.sql.schema import Identity, MetaData

# TODO Import this from typing when dropping Python 3.11
from typing_extensions import NotRequired, TypedDict, Unpack

from ixmp4 import db
from ixmp4.core.exceptions import Forbidden, IxmpError, ProgrammingError
from ixmp4.data import abstract, types
from ixmp4.data.abstract.annotations import (
    TabulateVersionsKwargs as TabulateVersionsKwargs,
)
from ixmp4.db import filters

if TYPE_CHECKING:
    from ixmp4.data.backend.db import SqlAlchemyBackend

logger = logging.getLogger(__name__)


@event.listens_for(Engine, "connect")
def set_sqlite_pragma(
    dbapi_connection: sqlite3.Connection, connection_record: Any
) -> None:
    if isinstance(dbapi_connection, sqlite3.Connection):
        cursor = dbapi_connection.cursor()
        cursor.execute("PRAGMA foreign_keys=ON")
        cursor.close()


@compiles(Identity, "sqlite")
def visit_identity(element: Any, compiler: Any, **kwargs: Any) -> TextClause:
    return text("")


class BaseModel(DeclarativeBase):
    NotFound: ClassVar[type[IxmpError]]
    NotUnique: ClassVar[type[IxmpError]]
    DeletionPrevented: ClassVar[type[IxmpError]]

    updateable_columns: ClassVar[list[str]] = []

    id: types.Id = db.Column(
        db.Integer,
        Identity(always=False, on_null=True, start=1, increment=1),
        primary_key=True,
        info={"skip_autogenerate": True},
    )

    def __str__(self) -> str:
        return self.__class__.__name__


BaseModel.metadata = MetaData(
    naming_convention={
        "ix": "ix_%(column_0_label)s",
        "uq": "uq_%(table_name)s_%(column_0_N_name)s",
        "ck": "ck_%(table_name)s_%(constraint_name)s",
        "fk": "fk_%(table_name)s_%(column_0_N_name)s_%(referred_table_name)s",
        "pk": "pk_%(table_name)s",
    }
)


SelectType = TypeVar("SelectType", bound=db.sql.Select[tuple[BaseModel, ...]])

ModelType = TypeVar("ModelType", bound=BaseModel)


class BaseRepository(Generic[ModelType]):
    backend: "SqlAlchemyBackend"
    session: Session
    dialect: Dialect
    bundle: Bundle[Any]
    model_class: type[ModelType]

    def __init__(self, backend: "SqlAlchemyBackend") -> None:
        self.backend = backend
        self.session = backend.session
        self.engine = backend.engine

        if self.session.bind is not None:
            self.dialect = self.session.bind.dialect
        else:
            raise ProgrammingError("Database session is closed.")

        self.bundle: Bundle[Any] = Bundle(
            self.model_class.__name__, *db.utils.get_columns(self.model_class).values()
        )

        super().__init__()


class Retriever(BaseRepository[ModelType], abstract.Retriever):
    def get(self, *args: Any, **kwargs: Any) -> ModelType:
        raise NotImplementedError


class CreateKwargs(TypedDict, total=False):
    dimension_id: int
    description: str
    name: str
    model_name: str
    scenario_name: str
    hierarchy: str
    run__id: int
    key: str
    value: abstract.annotations.PrimitiveTypes
    parameters: Mapping[str, Any]
    run_id: int
    unit_name: str | None
    column_names: list[str] | None
    constrained_to_indexsets: list[str] | None


class Creator(BaseRepository[ModelType], abstract.Creator):
    def add(self, *args: Any, **kwargs: Any) -> ModelType:
        raise NotImplementedError

    def create(self, *args: Any, **kwargs: Unpack[CreateKwargs]) -> ModelType:
        model = self.add(*args, **kwargs)
        try:
            self.session.commit()
        except IntegrityError:
            self.session.rollback()
            raise self.model_class.NotUnique(" | ".join([str(arg) for arg in args]))

        return model


class Deleter(BaseRepository[ModelType], abstract.Deleter):
    def delete(self, id: int) -> None:
        exc = db.select(self.model_class).where(self.model_class.id == id)

        try:
            obj = self.session.execute(exc).scalar_one()
        except NoResultFound:
            raise self.model_class.NotFound(id=id)

        try:
            self.session.delete(obj)
            self.session.commit()
        except IntegrityError:
            self.session.rollback()
            raise self.model_class.DeletionPrevented


class CheckAccessKwargs(TypedDict, total=False):
    run: abstract.annotations.HasRunFilter
    is_default: bool | None
    default_only: bool | None


class SelectCountKwargs(abstract.HasNameFilter, total=False):
    default_only: bool | None
    dimension_id: int | None
    dimension_id__in: Iterable[int] | None
    id__in: set[int]
    is_default: bool | None
    join_parameters: bool | None
    join_runs: bool | None
    join_run_id: bool | None
    iamc: abstract.annotations.IamcFilterAlias
    model: abstract.annotations.HasModelFilter
    region: abstract.annotations.HasRegionFilter
    run: abstract.annotations.HasRunFilter
    scenario: abstract.annotations.HasScenarioFilter
    unit: abstract.annotations.HasUnitFilter
    variable: abstract.annotations.HasVariableFilter


class Selecter(BaseRepository[ModelType]):
    filter_class: type[filters.BaseFilter] | None

    def check_access(
        self,
        ids: set[int],
        access_type: str = "view",
        **kwargs: Unpack[CheckAccessKwargs],
    ) -> None:
        exc = self.select_for_count(
            _exc=db.select(db.func.count()).select_from(self.model_class),
            id__in=ids,
            _access_type=access_type,
            **kwargs,
        )
        num_permitted_ids = self.session.execute(exc).scalar()
        num_ids = len(ids)
        if not num_permitted_ids == num_ids:
            logger.debug(
                f"Permission check failed {num_permitted_ids}/{num_ids} objects "
                "permitted."
            )
            raise Forbidden(f"Permission check failed for access type '{access_type}'.")

    SelectAuthType = TypeVar("SelectAuthType", bound=db.sql.Select[tuple[Any]])

    def join_auth(self, exc: SelectAuthType) -> SelectAuthType:
        return exc

    def apply_auth(self, exc: SelectAuthType, access_type: str) -> SelectAuthType:
        if self.backend.auth_context is not None:
            if not self.backend.auth_context.is_managed:
                exc = self.join_auth(exc)
            exc = self.backend.auth_context.apply(access_type, exc)
        return exc

    def select_for_count(
        self,
        _exc: db.sql.Select[tuple[int]],
        _filter: filters.BaseFilter | None = None,
        _skip_filter: bool = False,
        _access_type: str = "view",
        **kwargs: Unpack[SelectCountKwargs],
    ) -> db.sql.Select[tuple[int]]:
        if self.filter_class is None:
            cls_name = self.__class__.__name__
            raise NotImplementedError(
                f"Provide `{cls_name}.filter_class` or reimplement `{cls_name}.select`."
            )

        _exc = self.apply_auth(_exc, _access_type)
        if _filter is not None and not _skip_filter:
            filter_instance = _filter
            _exc = filter_instance.join(_exc, session=self.session)
            _exc = filter_instance.apply(_exc, self.model_class, self.session)
        elif not _skip_filter:
            kwarg_filter: filters.BaseFilter = self.filter_class(**kwargs)
            _exc = kwarg_filter.join(_exc, session=self.session)
            _exc = kwarg_filter.apply(_exc, self.model_class, self.session)
        return _exc

    def select(
        self,
        _filter: filters.BaseFilter | None = None,
        _exc: db.sql.Select[tuple[ModelType]] | None = None,
        _access_type: str = "view",
        _post_filter: Callable[
            [db.sql.Select[tuple[ModelType]]], db.sql.Select[tuple[ModelType]]
        ]
        | None = None,
        _skip_filter: bool = False,
        **kwargs: Any,
    ) -> db.sql.Select[tuple[ModelType]]:
        if self.filter_class is None:
            cls_name = self.__class__.__name__
            raise NotImplementedError(
                f"Provide `{cls_name}.filter_class` or reimplement `{cls_name}.select`."
            )

        if _exc is None:
            _exc = db.select(self.model_class)

        _exc = self.apply_auth(_exc, _access_type)

        if _filter is not None and not _skip_filter:
            filter_instance = _filter
            _exc = filter_instance.join(_exc, session=self.session)
            _exc = filter_instance.apply(_exc, self.model_class, self.session)
        elif not _skip_filter:
            kwarg_filter: filters.BaseFilter = self.filter_class(**kwargs)
            _exc = kwarg_filter.join(_exc, session=self.session)
            _exc = kwarg_filter.apply(_exc, self.model_class, self.session)

        if _post_filter is not None:
            _exc = _post_filter(_exc)
        return _exc


class QueryMixin(BaseRepository[ModelType]):
    def tabulate_query(self, exc: db.sql.Select[Any]) -> pd.DataFrame:
        if self.session.bind is not None:
            with self.engine.connect() as con:
                return pd.read_sql(exc, con=con).replace({np.nan: None})
        else:
            raise ProgrammingError("Database session is closed.")

    def list_query(self, exc: db.sql.Select[Any]) -> list[Any]:
        result = self.session.execute(exc).scalars().all()
        return list(result)


class Lister(QueryMixin[ModelType], Selecter[ModelType], abstract.Lister):
    def list(self, *args: Any, **kwargs: Any) -> list[ModelType]:
        _exc = self.select(*args, **kwargs)
        _exc = _exc.order_by(self.model_class.id.asc())
        return self.list_query(_exc)


class Tabulator(QueryMixin[ModelType], Selecter[ModelType], abstract.Tabulator):
    def tabulate(self, *args: Any, _raw: bool = False, **kwargs: Any) -> pd.DataFrame:
        _exc = self.select(*args, **kwargs)
        _exc = _exc.order_by(self.model_class.id.asc())

        return self.tabulate_query(_exc)


class PaginateKwargs(TypedDict):
    _filter: filters.BaseFilter
    join_parameters: NotRequired[bool | None]
    join_runs: NotRequired[bool | None]
    join_run_id: NotRequired[bool | None]
    join_run_index: NotRequired[bool | None]
    table: bool


class EnumerateKwargs(TypedDict):
    _filter: filters.BaseFilter
    join_parameters: NotRequired[bool | None]
    join_runs: NotRequired[bool | None]
    join_run_id: NotRequired[bool | None]
    join_run_index: NotRequired[bool | None]
    _post_filter: Callable[
        [db.sql.Select[tuple[ModelType]]], db.sql.Select[tuple[ModelType]]
    ]


class CountKwargs(abstract.HasNameFilter, total=False):
    dimension_id: int | None
    dimension_id__in: Iterable[int] | None
    _filter: filters.BaseFilter
    join_parameters: bool | None
    join_runs: bool | None
    join_run_id: bool | None
    iamc: abstract.annotations.IamcFilterAlias
    model: abstract.annotations.HasModelFilter
    region: abstract.annotations.HasRegionFilter
    run: abstract.annotations.HasRunFilter
    scenario: abstract.annotations.HasScenarioFilter
    unit: abstract.annotations.HasUnitFilter
    variable: abstract.annotations.HasVariableFilter


class Enumerator(Lister[ModelType], Tabulator[ModelType], abstract.Enumerator):
    def enumerate(
        self, table: bool = False, **kwargs: Unpack[EnumerateKwargs]
    ) -> list[ModelType] | pd.DataFrame:
        return self.tabulate(**kwargs) if table else self.list(**kwargs)

    def paginate(
        self, limit: int = 1000, offset: int = 0, **kwargs: Unpack[PaginateKwargs]
    ) -> list[ModelType] | pd.DataFrame:
        return self.enumerate(
            **kwargs, _post_filter=lambda e: e.offset(offset).limit(limit)
        )

    def count(self, **kwargs: Unpack[CountKwargs]) -> int:
        _exc = self.select_for_count(
            _exc=db.select(db.func.count(self.model_class.id.distinct())),
            **kwargs,
        )
        return self.session.execute(_exc).scalar_one()


class TransactionProtocol(Protocol):
    """Protocol class to mock the type of sqlalchemy_continuums"""

    id: int
    issued_at: datetime


class BulkOperator(Tabulator[ModelType]):
    merge_suffix: str = "_y"

    @property
    def max_list_length(self) -> int:
        return 50_000

    def merge_existing(
        self, df: pd.DataFrame, existing_df: pd.DataFrame
    ) -> pd.DataFrame:
        columns = db.utils.get_columns(self.model_class)
        primary_key_columns = db.utils.get_pk_columns(self.model_class)
        on = (
            (
                set(existing_df.columns) & set(df.columns) & set(columns.keys())
            )  # all cols which exist in both dfs and the db model
            - set(self.model_class.updateable_columns)  # no updateable columns
            - set(primary_key_columns.keys())  # no pk columns
        )  # = all columns that are constant and provided during creation

        for x in existing_df.select_dtypes(include=["datetime64"]).columns.tolist():
            existing_df.loc[:, x] = existing_df[x].astype(object)

        for x in df.select_dtypes(include=["datetime64"]).columns.tolist():
            df.loc[:, x] = df[x].astype(object)

        return df.merge(
            existing_df,
            how="left",
            on=list(on),
            suffixes=(None, self.merge_suffix),
        )

    def drop_merge_artifacts(
        self, df: pd.DataFrame, extra_columns: list[str] | None = None
    ) -> pd.DataFrame:
        if extra_columns is None:
            extra_columns = []

        existing_columns = [
            str(col) for col in df.columns if col.endswith(self.merge_suffix)
        ]

        df = df.drop(columns=existing_columns + extra_columns)
        df = df.dropna(axis="index", how="all")
        df = df.replace({np.nan: None})

        return df

    def split_by_max_unique_values(
        self, df: pd.DataFrame, columns: Iterable[str], mu: int
    ) -> tuple[pd.DataFrame, pd.DataFrame]:
        """
        Splits a DataFrame into two parts based on the maximum allowed unique
        values in specified columns.

        The function iteratively adjusts the chunk size starting from
        the entire DataFrame, reducing it until the maximum number of unique
        values in any of the specified columns within the chunk is ≤ `mu`.
        The remaining rows form the second part.

        Args:
            df (pd.DataFrame): The input DataFrame to split.
            columns (Iterable[str]): Columns to check for the maximum unique values.
            mu (int): Maximum allowed unique values in any
                of the specified columns for the first chunk.

        Returns:
            tuple[pd.DataFrame, pd.DataFrame]:
                - First element (chunk_df): Largest initial chunk where
                    columns have ≤ `mu` unique values.
                - Second element (remaining_df): Remaining rows of the
                    DataFrame after `chunk_df`.

        Example:
            If `df` has 100 rows and a column initially has 10 unique values
            (with `mu=5`), the chunk is reduced to 50 rows.
            If the chunk then has 6 unique values, it's further
            reduced to ~41 rows, and so on until all columns in the chunk
            meet the condition.
        """

        chunk_size = len(df.index)
        remaining_df = pd.DataFrame()

        if chunk_size <= mu:
            return df, remaining_df

        max_ = chunk_size
        chunk_df = df
        while True:
            max_ = max(chunk_df[c].nunique() for c in columns)
            if max_ <= mu:
                break
            chunk_size = int(np.floor((mu / max_) * chunk_size))
            chunk_df = df.iloc[:chunk_size, :]
            remaining_df = df.iloc[chunk_size:, :]

        return chunk_df, remaining_df

    def tabulate_existing(self, df: pd.DataFrame) -> pd.DataFrame:
        exc = db.select(self.model_class)
        foreign_columns = db.utils.get_foreign_columns(self.model_class)

        for col in foreign_columns:
            foreign_pks = df[col.name].unique().tolist()
            exc = exc.where(col.in_(foreign_pks))
        return self.tabulate(_exc=exc, _raw=True, _skip_filter=True)

    def yield_chunks(self, df: pd.DataFrame) -> Iterator[pd.DataFrame]:
        foreign_columns = db.utils.get_foreign_columns(self.model_class)
        foreign_names = [c.name for c in foreign_columns]
        remaining_df = df.sort_values(foreign_names)

        while not remaining_df.empty:
            chunk_df, remaining_df = self.split_by_max_unique_values(
                pd.DataFrame(remaining_df), foreign_names, self.max_list_length
            )
            yield pd.DataFrame(chunk_df)


class BulkUpserter(BulkOperator[ModelType]):
    def bulk_upsert(self, df: pd.DataFrame) -> None:
        if len(df.index) < self.max_list_length:
            self.bulk_upsert_chunk(df)
        else:
            for chunk_df in self.yield_chunks(df):
                self.bulk_upsert_chunk(pd.DataFrame(chunk_df))

    def bulk_upsert_chunk(self, df: pd.DataFrame) -> None:
        logger.debug(f"Starting `bulk_upsert_chunk` for {len(df)} rows.")
        columns = db.utils.get_columns(self.model_class)
        df = df[list(set(columns.keys()) & set(df.columns))].copy()
        existing_df = self.tabulate_existing(df)
        if existing_df.empty:
            logger.debug(f"Inserting {len(df)} rows.")
            self.bulk_insert(df, skip_validation=True)
        else:
            df = self.merge_existing(df, existing_df)
            df["exists"] = np.where(pd.notnull(df["id"]), True, False)
            cond = []
            for col in self.model_class.updateable_columns:
                updated_col = col + self.merge_suffix
                if updated_col in df.columns:
                    # coerce to same type so the inequality
                    # operation works with pyarrow installed
                    df[updated_col] = df[updated_col].astype(df[col].dtype)
                    are_not_equal = df[col] != df[updated_col]
                    # extra check if both values are NA because NA == NA = NA
                    # in pandas with pyarrow
                    both_are_na = pd.isna(df[col]) & pd.isna(df[updated_col])
                    cond.append(~both_are_na | are_not_equal)

            df["differs"] = np.where(np.logical_or.reduce(cond), True, False)

            insert_df = self.drop_merge_artifacts(
                df.where(~df["exists"]), extra_columns=["id", "differs", "exists"]
            )
            update_df = (
                self.drop_merge_artifacts(  # gets rid of metas null columns, fix!
                    df.where(df["exists"] & df["differs"]),
                    extra_columns=["differs", "exists"],
                )
            )

            if not insert_df.empty:
                logger.debug(f"Inserting {len(insert_df)} rows.")
                self.bulk_insert(insert_df, skip_validation=True)
            if not update_df.empty:
                logger.debug(f"Updating {len(update_df)} rows.")
                self.bulk_update(update_df, skip_validation=True)

        self.session.commit()

    def bulk_insert(self, df: pd.DataFrame, **kwargs: Any) -> None:
        if "id" in df.columns:
            raise ProgrammingError("You may not insert the 'id' column.")

        # to_dict returns a more general list[Mapping[Hashable, Unknown]]
        m = cast(list[dict[str, Any]], df.to_dict("records"))

        exc = db.insert(self.model_class).execution_options(synchronize_session=False)

        try:
            self.session.execute(exc, m)
        except IntegrityError as e:
            raise self.model_class.NotUnique(*e.args)

    def bulk_update(self, df: pd.DataFrame, **kwargs: Any) -> None:
        # to_dict returns a more general list[Mapping[Hashable, Unknown]]
        m = cast(list[dict[str, Any]], df.to_dict("records"))

        self.session.execute(
            db.update(self.model_class).execution_options(synchronize_session=False), m
        )


class BulkDeleter(BulkOperator[ModelType]):
    def bulk_delete(self, df: pd.DataFrame) -> None:
        for chunk_df in self.yield_chunks(df):
            self.bulk_delete_chunk(chunk_df)

    def bulk_delete_chunk(self, df: pd.DataFrame) -> None:
        existing_df = self.tabulate_existing(df)
        if existing_df.empty:
            return

        df = self.merge_existing(df, existing_df)
        df["exists"] = np.where(pd.notnull(df["id"]), True, False)
        delete_df = df.where(df["exists"])
        id_df = delete_df[["id"]]

        excs = []
        for _, sdf in id_df.groupby(id_df.index // self.max_list_length):
            dexc = db.delete(self.model_class).where(self.model_class.id.in_(sdf["id"]))
            excs.append(dexc)

        for exc in excs:
            self.session.execute(exc, execution_options={"synchronize_session": False})

        self.session.commit()


class TabulateTransactionsKwargs(abstract.annotations.HasPaginationArgs, total=False):
    pass



from typing import ClassVar

from ixmp4.db import Session, filters, sql

from .. import RunMetaEntry


class RunMetaEntryFilter(filters.BaseFilter, metaclass=filters.FilterMeta):
    sqla_model: ClassVar[type] = RunMetaEntry

    id: filters.Id
    dtype: filters.String
    run__id: filters.Integer | None = filters.Field(None, alias="run_id")

    key: filters.String

    value_int: filters.Integer
    value_str: filters.String
    value_float: filters.Float
    value_bool: filters.Boolean

    # NOTE specific type hint here is based on usage; adapt accordingly
    def join(
        self, exc: sql.Select[tuple[RunMetaEntry]], session: Session | None = None
    ) -> sql.Select[tuple[RunMetaEntry]]:
        return exc



from typing import Any

from ixmp4.data.db.docs import BaseDocsRepository, docs_model

from .model import OptimizationVariable as Variable

OptimizationVariableDocs = docs_model(Variable)


class OptimizationVariableDocsRepository(BaseDocsRepository[Any]):
    model_class = OptimizationVariableDocs
    dimension_model_class = Variable



from typing import Any

from ..docs import BaseDocsRepository, docs_model
from .model import Scenario

ScenarioDocs = docs_model(Scenario)


class ScenarioDocsRepository(BaseDocsRepository[Any]):
    model_class = ScenarioDocs
    dimension_model_class = Scenario



# type: ignore
"""Introduce optimization item delete()

Revision ID: 1e5554a164e6
Revises: 8fd96526b04f
Create Date: 2025-05-08 12:13:39.340523

"""

from alembic import op

# Revision identifiers, used by Alembic.
revision = "1e5554a164e6"
down_revision = "8fd96526b04f"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table(
        "optimization_equationindexsetassociation", schema=None
    ) as batch_op:
        batch_op.drop_constraint(
            "fk_optimization_equationindexsetassociation_equation__i_be8b",
            type_="foreignkey",
        )
        batch_op.create_foreign_key(
            batch_op.f("fk_optimization_equationindexsetassociation_equation__i_be8b"),
            "optimization_equation",
            ["equation__id"],
            ["id"],
            ondelete="CASCADE",
        )

    with op.batch_alter_table("optimization_indexsetdata", schema=None) as batch_op:
        batch_op.drop_constraint(
            "fk_optimization_indexsetdata_indexset__id_optimization_indexset",
            type_="foreignkey",
        )
        batch_op.create_foreign_key(
            batch_op.f(
                "fk_optimization_indexsetdata_indexset__id_optimization_indexset"
            ),
            "optimization_indexset",
            ["indexset__id"],
            ["id"],
            ondelete="CASCADE",
        )

    with op.batch_alter_table(
        "optimization_parameterindexsetassociation", schema=None
    ) as batch_op:
        batch_op.drop_constraint(
            "fk_optimization_parameterindexsetassociation_parameter__8a6d",
            type_="foreignkey",
        )
        batch_op.create_foreign_key(
            batch_op.f("fk_optimization_parameterindexsetassociation_parameter__8a6d"),
            "optimization_parameter",
            ["parameter__id"],
            ["id"],
            ondelete="CASCADE",
        )

    with op.batch_alter_table(
        "optimization_tableindexsetassociation", schema=None
    ) as batch_op:
        batch_op.drop_constraint(
            "fk_optimization_tableindexsetassociation_table__id_opti_5cfe",
            type_="foreignkey",
        )
        batch_op.create_foreign_key(
            batch_op.f("fk_optimization_tableindexsetassociation_table__id_opti_5cfe"),
            "optimization_table",
            ["table__id"],
            ["id"],
            ondelete="CASCADE",
        )

    with op.batch_alter_table(
        "optimization_variableindexsetassociation", schema=None
    ) as batch_op:
        batch_op.drop_constraint(
            "fk_optimization_variableindexsetassociation_variable__i_02a8",
            type_="foreignkey",
        )
        batch_op.create_foreign_key(
            batch_op.f("fk_optimization_variableindexsetassociation_variable__i_02a8"),
            "optimization_optimizationvariable",
            ["variable__id"],
            ["id"],
            ondelete="CASCADE",
        )

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table(
        "optimization_variableindexsetassociation", schema=None
    ) as batch_op:
        batch_op.drop_constraint(
            batch_op.f("fk_optimization_variableindexsetassociation_variable__i_02a8"),
            type_="foreignkey",
        )
        batch_op.create_foreign_key(
            "fk_optimization_variableindexsetassociation_variable__i_02a8",
            "optimization_optimizationvariable",
            ["variable__id"],
            ["id"],
        )

    with op.batch_alter_table(
        "optimization_tableindexsetassociation", schema=None
    ) as batch_op:
        batch_op.drop_constraint(
            batch_op.f("fk_optimization_tableindexsetassociation_table__id_opti_5cfe"),
            type_="foreignkey",
        )
        batch_op.create_foreign_key(
            "fk_optimization_tableindexsetassociation_table__id_opti_5cfe",
            "optimization_table",
            ["table__id"],
            ["id"],
        )

    with op.batch_alter_table(
        "optimization_parameterindexsetassociation", schema=None
    ) as batch_op:
        batch_op.drop_constraint(
            batch_op.f("fk_optimization_parameterindexsetassociation_parameter__8a6d"),
            type_="foreignkey",
        )
        batch_op.create_foreign_key(
            "fk_optimization_parameterindexsetassociation_parameter__8a6d",
            "optimization_parameter",
            ["parameter__id"],
            ["id"],
        )

    with op.batch_alter_table("optimization_indexsetdata", schema=None) as batch_op:
        batch_op.drop_constraint(
            batch_op.f(
                "fk_optimization_indexsetdata_indexset__id_optimization_indexset"
            ),
            type_="foreignkey",
        )
        batch_op.create_foreign_key(
            "fk_optimization_indexsetdata_indexset__id_optimization_indexset",
            "optimization_indexset",
            ["indexset__id"],
            ["id"],
        )

    with op.batch_alter_table(
        "optimization_equationindexsetassociation", schema=None
    ) as batch_op:
        batch_op.drop_constraint(
            batch_op.f("fk_optimization_equationindexsetassociation_equation__i_be8b"),
            type_="foreignkey",
        )
        batch_op.create_foreign_key(
            "fk_optimization_equationindexsetassociation_equation__i_be8b",
            "optimization_equation",
            ["equation__id"],
            ["id"],
        )

    # ### end Alembic commands ###
